from __future__ import annotations

import re

URL_PATTERN = re.compile(r"(?P<url>(?:https?://|www\.)[^\s<>\"]+)", re.IGNORECASE)
ALNUM_CHUNK_PATTERN = re.compile(r"[A-Za-z0-9]+|[^A-Za-z0-9]")
MIXED_CHUNK_PATTERN = re.compile(r"\d+|[A-Za-z]+")

URL_SEPARATOR_WORDS = {
    ":": "двоеточие",
    "/": "слэш",
    ".": "точка",
    "?": "вопрос",
    "&": "амперсанд",
    "=": "равно",
    "-": "дефис",
    "_": "нижнее подчёркивание",
    "#": "решётка",
    "%": "процент",
    "+": "плюс",
    "@": "собака",
    "~": "тильда",
}

DIGIT_WORDS = {
    "0": "ноль",
    "1": "один",
    "2": "два",
    "3": "три",
    "4": "четыре",
    "5": "пять",
    "6": "шесть",
    "7": "семь",
    "8": "восемь",
    "9": "девять",
}

TRAILING_URL_PUNCTUATION = ".,;:!?"
UNBALANCED_CLOSERS = {
    ")": "(",
    "]": "[",
    "}": "{",
}


def _split_trailing_punctuation(url: str) -> tuple[str, str]:
    suffix: list[str] = []

    while url and url[-1] in TRAILING_URL_PUNCTUATION:
        suffix.append(url[-1])
        url = url[:-1]

    while url and url[-1] in UNBALANCED_CLOSERS:
        closer = url[-1]
        opener = UNBALANCED_CLOSERS[closer]
        if url.count(closer) > url.count(opener):
            suffix.append(closer)
            url = url[:-1]
            continue
        break

    return url, "".join(reversed(suffix))


def _read_digit_run(chunk: str) -> str:
    return " ".join(DIGIT_WORDS[digit] for digit in chunk)


def _normalize_alnum_chunk(chunk: str) -> str:
    if chunk.isdigit():
        return _read_digit_run(chunk)
    if chunk.isalpha():
        return chunk

    pieces: list[str] = []
    for part in MIXED_CHUNK_PATTERN.findall(chunk):
        if part.isdigit():
            pieces.append(_read_digit_run(part))
        else:
            pieces.append(part)
    return " ".join(piece for piece in pieces if piece)


def _normalize_url(url: str) -> str:
    pieces: list[str] = []
    for chunk in ALNUM_CHUNK_PATTERN.findall(url):
        if not chunk:
            continue
        if chunk.isalnum():
            pieces.append(_normalize_alnum_chunk(chunk))
            continue
        for char in chunk:
            pieces.append(URL_SEPARATOR_WORDS.get(char, char))
    return re.sub(r"\s+", " ", " ".join(piece for piece in pieces if piece)).strip()


def normalize_urls(text: str, *, enabled: bool) -> str:
    if not enabled or not URL_PATTERN.search(text):
        return text

    def repl(match: re.Match[str]) -> str:
        raw_url = match.group("url")
        url, suffix = _split_trailing_punctuation(raw_url)
        if not url:
            return raw_url
        return f"{_normalize_url(url)}{suffix}"

    return URL_PATTERN.sub(repl, text)



from __future__ import annotations

import re
from pathlib import Path

from ._morph import get_morph
from .abbreviations import expand_abbreviations
from .caps import (
    normalize_caps_lines,
    normalize_first_word_caps,
    normalize_sentence_start_caps,
)
from .dates_time import normalize_dates_and_time
from .dictionary import apply_dictionary_rules
from .latinization import DEFAULT_DICTIONARIES_PATH, apply_latinization
from .numbering import convert_bracketed_numbers, convert_line_numbering
from .numerals import (
    ALL_UNITS,
    normalize_decimals,
    normalize_fractions,
    normalize_hyphenated_words,
    normalize_math_symbols,
    normalize_numerals,
    normalize_ordinals,
)
from .numerals._hyphen import (
    normalize_numeric_unit_hyphen_links,
    normalize_spaced_numeric_hyphen_words,
)
from .options import NormalizeOptions
from .preprocess_utils import (
    SLASH_FIX_PATTERN,
    UNIT_SLASH_PLACEHOLDER,
    apply_cleanup_replacements,
    clean_numbers,
    expand_years_ago_abbreviation,
    normalize_ascii_quote_pairs,
    normalize_cyrillic_combining_stress_marks,
    normalize_explicit_dashes,
    normalize_linebreaks,
    normalize_numeric_abbreviations,
    normalize_punctuation_spacing,
    normalize_spaced_ascii_hyphens,
    normalize_unicode_fractions,
    protect_letter_hyphens,
    protect_negative_numbers,
    protect_unit_slashes,
    remove_decorative_separators,
    remove_numeric_footnotes,
    remove_zero_width_formatting,
    restore_letter_hyphens,
    restore_paragraph_breaks,
)
from .roman_numerals import normalize_roman
from .urls import normalize_urls
from .years import normalize_numeric_ranges, normalize_years

PARTICLE_PATTERN = re.compile(
    r"(?<=[а-яА-ЯёЁ])\s*[–—―]\s*(?=(?:то|либо|нибудь|ка|таки|де|с)\b)",
    re.IGNORECASE,
)
DECORATIVE_MARKER_PATTERN = re.compile(r"(?:[*=_~+#\-\xad\u2010-\u2015]\s*){5,}")
ASTERISK_SEPARATOR_PATTERN = re.compile(r"(?:\*\s*){3,}")
CURRENCY_AMOUNT_PATTERN = re.compile(r"([$€₽])\s?(\d+(?:[.,]\d+)?)")
DEGREE_SPACING_PATTERN = re.compile(r"(\d)°")

GLUED_PREPOSITIONS = {
    "до",
    "по",
    "от",
    "из",
    "с",
    "со",
    "к",
    "ко",
    "в",
    "во",
    "на",
    "у",
    "о",
    "об",
    "обо",
    "при",
    "за",
    "под",
    "над",
    "без",
    "для",
    "ради",
    "через",
    "сквозь",
    "вдоль",
    "мимо",
    "кроме",
    "вместо",
    "против",
    "среди",
    "между",
    "перед",
    "пред",
    "близ",
    "около",
    "свыше",
}


class PipelineNormalizer:
    """Fixed-order Russian normalization pipeline with toggleable stages."""

    def __init__(self, options: NormalizeOptions | None = None) -> None:
        self.options = options or NormalizeOptions()

    @property
    def dictionaries_path(self) -> Path:
        return self.options.dictionaries_path or DEFAULT_DICTIONARIES_PATH

    def run_stage(self, stage: str, text: str) -> str:
        handlers = {
            "urls": self.run_urls,
            "preprocess": self.run_preprocess,
            "roman": self.run_roman,
            "years": self.run_years,
            "dates_time": self.run_dates_time,
            "numerals": self.run_numerals,
            "abbreviations": self.run_abbreviations,
            "dictionary": self.run_dictionary,
            "latinization": self.run_latinization,
            "finalize": self.run_finalize,
        }
        try:
            handler = handlers[stage]
        except KeyError as exc:
            available = ", ".join(sorted(handlers))
            raise ValueError(
                f"Unknown stage '{stage}'. Available stages: {available}"
            ) from exc
        return handler(text)

    def normalize_text(self, text: str) -> str:
        text = self.run_urls(text)
        text = self._run_preprocess_steps(
            text,
            keep_paragraph_placeholders=True,
            apply_caps_normalization=False,
        )
        text = self.run_roman(text)
        text = self._run_caps_normalization(text)
        if self.options.remove_links:
            text = remove_numeric_footnotes(
                text,
                keep_paragraph_placeholders=True,
                ignore_interval=self.options.remove_links_ignore_interval,
            )
        else:
            text = normalize_linebreaks(text, keep_paragraph_placeholders=True).strip()
        text = self.run_years(text)
        text = self.run_dates_time(text)
        text = self.run_numerals(text)
        text = self.run_abbreviations(text)
        text = self.run_latinization(text)
        text = self.run_dictionary(text)
        return self.run_finalize(text)

    def run_preprocess(
        self, text: str, keep_paragraph_placeholders: bool = False
    ) -> str:
        text = self._run_preprocess_steps(
            text,
            keep_paragraph_placeholders=keep_paragraph_placeholders,
            apply_caps_normalization=True,
        )
        if self.options.remove_links:
            text = remove_numeric_footnotes(
                text,
                keep_paragraph_placeholders=keep_paragraph_placeholders,
                ignore_interval=self.options.remove_links_ignore_interval,
            )
        else:
            text = normalize_linebreaks(
                text, keep_paragraph_placeholders=keep_paragraph_placeholders
            ).strip()
        return text

    def run_urls(self, text: str) -> str:
        return normalize_urls(text, enabled=self.options.enable_url_normalization)

    def _run_preprocess_steps(
        self,
        text: str,
        *,
        keep_paragraph_placeholders: bool,
        apply_caps_normalization: bool,
    ) -> str:
        if text.strip().startswith("+"):
            text = " " + text.lstrip()

        text = normalize_linebreaks(
            text, keep_paragraph_placeholders=keep_paragraph_placeholders
        )
        text = remove_zero_width_formatting(text)
        text = normalize_cyrillic_combining_stress_marks(text)
        text = protect_letter_hyphens(text)
        text = text.replace("◦", " ")
        text = remove_decorative_separators(text)
        text = DECORATIVE_MARKER_PATTERN.sub(" ", text)
        text = ASTERISK_SEPARATOR_PATTERN.sub(" ", text)
        text = text.replace("[", "(").replace("]", ")")
        text = protect_unit_slashes(text)
        text = SLASH_FIX_PATTERN.sub(" ", text)
        text = text.replace(UNIT_SLASH_PLACEHOLDER, "/")
        text = normalize_ascii_quote_pairs(text)
        text = normalize_punctuation_spacing(text)
        text = expand_years_ago_abbreviation(text)
        text = normalize_numeric_abbreviations(text)
        text = protect_negative_numbers(text)
        text = normalize_explicit_dashes(text)
        text = convert_bracketed_numbers(text, self.options)
        text = convert_line_numbering(text)
        if apply_caps_normalization:
            text = self._run_caps_normalization(text)
        text = PARTICLE_PATTERN.sub("-", text)
        text = apply_cleanup_replacements(text)
        text = normalize_unicode_fractions(text)
        text = text.translate(
            str.maketrans("⁰¹²³⁴⁵⁶⁷⁸⁹₀₁₂₃₄₅₆₇₈₉", "01234567890123456789")
        )
        text = CURRENCY_AMOUNT_PATTERN.sub(r"\2\1", text)
        text = DEGREE_SPACING_PATTERN.sub(r"\1 °", text)
        text = clean_numbers(text)
        text = self._fix_glued_numbers(text)
        text = restore_letter_hyphens(text)
        return text

    def _run_caps_normalization(self, text: str) -> str:
        text = normalize_caps_lines(text, enabled=self.options.enable_caps_normalization)
        return normalize_first_word_caps(
            text, enabled=self.options.enable_first_word_decap
        )

    def run_roman(self, text: str) -> str:
        return normalize_roman(text, self.options)

    def run_years(self, text: str) -> str:
        text = normalize_years(text, self.options)
        return normalize_numeric_ranges(text)

    def run_dates_time(self, text: str) -> str:
        if not self.options.enable_dates_time_normalization:
            return text
        return normalize_dates_and_time(text, self.options)

    def run_numerals(self, text: str) -> str:
        if not self.options.enable_numeral_normalization:
            return text
        text = normalize_math_symbols(text)
        text = normalize_spaced_numeric_hyphen_words(text)
        text = normalize_numeric_unit_hyphen_links(text)
        text = normalize_decimals(text)
        text = normalize_fractions(text)
        text = normalize_hyphenated_words(text)
        text = normalize_ordinals(text)
        return normalize_numerals(text, self.options)

    def run_abbreviations(self, text: str) -> str:
        return expand_abbreviations(text, self.options)

    def run_dictionary(self, text: str) -> str:
        return apply_dictionary_rules(
            text,
            enabled=self.options.enable_dictionary_normalization,
            dictionaries_path=self.dictionaries_path,
            include_only_files=self.options.dictionary_include_files,
            exclude_files=self.options.dictionary_exclude_files,
        )

    def run_latinization(self, text: str) -> str:
        return apply_latinization(
            text,
            enabled=self.options.enable_latinization,
            backend=self.options.latinization_backend,
            dictionaries_path=self.dictionaries_path,
            dictionary_filename=self.options.latin_dictionary_filename,
            include_stress_markers=self.options.enable_latinization_stress_marks,
        )

    def run_finalize(self, text: str) -> str:
        text = normalize_ascii_quote_pairs(text)
        text = normalize_spaced_ascii_hyphens(text)
        text = normalize_punctuation_spacing(text)
        text = normalize_linebreaks(text, keep_paragraph_placeholders=True)
        text = normalize_sentence_start_caps(text)
        return restore_paragraph_breaks(text)

    def _fix_glued_numbers(self, text: str) -> str:
        def fix_glued(match: re.Match[str]) -> str:
            num, word = match.group(1), match.group(2)
            word_lower = word.lower()
            if word == "k":
                return f"{num} тыс"
            if word_lower in {
                "ти",
                "ми",
                "го",
                "му",
                "м",
                "х",
                "я",
                "е",
                "й",
                "о",
                "а",
            }:
                return f"{num}-{word}"
            if word_lower in ALL_UNITS or word_lower in GLUED_PREPOSITIONS:
                return f"{num} {word}"

            parsed = get_morph().parse(word_lower)[0]
            if "ADJF" in parsed.tag:
                return f"{num}-{word}"
            if "NOUN" in parsed.tag:
                return f"{num} {word}"
            return f"{num}-{word}"

        text = re.sub(r"(?<=[а-яА-ЯёЁa-zA-Z])(\d+)", r" \1", text)
        previous = None
        iteration = 0
        while previous != text and iteration < 5:
            previous = text
            text = re.sub(r"(\d+)([а-яА-ЯёЁa-zA-Z+]{1,})", fix_glued, text)
            iteration += 1
        return text


def preprocess_text(text: str, options: NormalizeOptions | None = None) -> str:
    return PipelineNormalizer(options).run_preprocess(text)



from __future__ import annotations

import functools
import re

import num2words

from ._morph import get_morph
from .options import NormalizeOptions
from .text_context import normalize_context_token, simple_tokenize
from .years_context import (
    YEAR_ANY_NUMBER_PATTERN,
    YEAR_IMPLICIT_PREP_PATTERN,
    YEAR_RANGE_NUMBER_PATTERN,
    is_plausible_year,
    should_treat_as_implicit_year,
)

YEAR_SUFFIX_TO_CASE = {
    "ый": "nomn",
    "ой": "nomn",
    "й": "nomn",
    "ые": "nomn",
    "ого": "gent",
    "го": "gent",
    "ых": "gent",
    "ому": "datv",
    "ым": "ablt",
    "ыми": "ablt",
    "ом": "loct",
    "му": "datv",
    "е": "nomn",
    "х": "gent",
    "м": "datv",
    "ми": "ablt",
}
YEAR_SUFFIX_REGEX = "|".join(
    sorted((re.escape(suffix) for suffix in YEAR_SUFFIX_TO_CASE), key=len, reverse=True)
)
PLURAL_SUFFIXES = {"е", "х", "м", "ми", "ые", "ых", "ыми"}
YEAR_WORD_TO_CASE = {
    "год": "nomn",
    "года": "gent",
    "году": "loct",
    "годом": "ablt",
    "годе": "loct",
    "годы": "nomn",
    "годов": "gent",
    "годам": "datv",
    "года́ми": "ablt",
    "годами": "ablt",
    "годах": "loct",
    "г.": "nomn",
    "гг.": "nomn",
}
YEAR_WORD_FORMS = {
    ("год", "nomn"): "год",
    ("год", "gent"): "года",
    ("год", "datv"): "году",
    ("год", "accs"): "год",
    ("год", "ablt"): "годом",
    ("год", "loct"): "году",
    ("годы", "nomn"): "годы",
    ("годы", "gent"): "годов",
    ("годы", "datv"): "годам",
    ("годы", "accs"): "годы",
    ("годы", "ablt"): "годами",
    ("годы", "loct"): "годах",
}
YEAR_PLURAL_ABBREV_REGEX = r"(?:гг\.?|г\.\s*г\.?)"
PREPOSITIONS_TO_CASE = {
    "в": "loct",
    "во": "loct",
    "о": "loct",
    "об": "loct",
    "к": "datv",
    "ко": "datv",
    "с": "gent",
    "со": "gent",
    "до": "gent",
    "от": "gent",
    "за": "accs",
    "на": "accs",
    "по": "accs",
    "между": "ablt",
    "около": "gent",
}
NUMERIC_RANGE_PATTERN = re.compile(
    r"(\d+)\s*[—–]\s*(\d+)(?!\d)(?!\s*[а-яА-ЯёЁa-zA-Z%°$€₽Ω])"
)
YEAR_PREPOSITION_REGEX = r"в|во|о|об|к|ко|с|со|до|от|за|на|по|между|около"
YEAR_IMPLICIT_PREP_REGEX = r"в|во|о|об|к|ко|на|за|под"
ERA_REGEX = (
    r"(?:до\s+н\.?\s*э\.?|н\.?\s*э\.?|до\s+нашей\s+эры|нашей\s+эры|"
    r"B\.?\s*C\.?\s*E\.?|B\.?\s*C\.?|C\.?\s*E\.?|A\.?\s*D\.?)"
)


def _normalize_era_marker(raw: str) -> str | None:
    compact = re.sub(r"[\s.]+", "", raw)
    compact_upper = compact.upper()
    if compact_upper in {"BC", "BCE"} and compact == compact_upper:
        return "до нашей эры"
    if compact_upper in {"AD", "CE"} and compact == compact_upper:
        return "нашей эры"

    lowered = raw.lower().strip()
    if re.fullmatch(r"до\s+н\.?\s*э\.?", lowered, re.IGNORECASE):
        return "до нашей эры"
    if re.fullmatch(r"н\.?\s*э\.?", lowered, re.IGNORECASE):
        return "нашей эры"
    if lowered.startswith("до нашей эры"):
        return "до нашей эры"
    if lowered.startswith("нашей эры"):
        return "нашей эры"
    return None


def _should_keep_era_terminal_dot(source_text: str, match: re.Match[str]) -> bool:
    raw_era = match.group("era")
    if raw_era is None or not raw_era.rstrip().endswith("."):
        return False

    rest = source_text[match.end() :]
    stripped_rest = rest.lstrip()
    if not stripped_rest:
        return True
    if "\n" in rest[: len(rest) - len(stripped_rest)]:
        return True
    next_char = stripped_rest[:1]
    if next_char in '.!?…)"»”]}':
        return True
    if next_char.isupper():
        return True
    if (
        next_char in '"»”)]}'
        and len(stripped_rest) > 1
        and stripped_rest[1:2].isupper()
    ):
        return True
    return False


def _resolve_explicit_year_word_case(prep: str | None, word: str) -> str:
    word_lower = word.lower()
    if word_lower == "году" and prep:
        return "datv" if prep.lower() in ("к", "ко") else "loct"
    return YEAR_WORD_TO_CASE.get(word_lower, "nomn")


@functools.lru_cache(maxsize=1024)
def year_to_ordinal_words(year: int, case: str = "nomn", plural: bool = False) -> str:
    normalized_case = "nomn" if case == "accs" and not plural else case
    cases_map = {
        "nomn": "nominative",
        "gent": "genitive",
        "datv": "dative",
        "accs": "accusative",
        "ablt": "instrumental",
        "loct": "prepositional",
    }
    if year >= 10000 and not plural and year % 1000:
        high = (year // 1000) * 1000
        tail = year % 1000
        try:
            high_words = num2words.num2words(
                high,
                lang="ru",
                case=cases_map.get(normalized_case, "nominative"),
            )
            tail_words = year_to_ordinal_words(tail, normalized_case)
            return f"{high_words} {tail_words}"
        except Exception:
            pass
    if normalized_case in cases_map:
        try:
            return num2words.num2words(
                year,
                lang="ru",
                to="ordinal",
                case=cases_map[normalized_case],
                plural=plural,
            )
        except Exception:
            pass
    try:
        ordinal = num2words.num2words(year, lang="ru", to="ordinal")
    except Exception:
        return str(year)
    words = ordinal.split()
    if not words:
        return ordinal
    parsed = get_morph().parse(words[-1])
    if not parsed:
        return ordinal
    tags = {normalized_case, "plur" if plural else "masc"}
    if "Anum" in parsed[0].tag:
        tags.add("Anum")
    inflected = parsed[0].inflect(tags)
    if inflected:
        words[-1] = inflected.word
    return " ".join(words)


def normalize_numeric_ranges(text: str) -> str:
    return NUMERIC_RANGE_PATTERN.sub(lambda m: f"{m.group(1)}, {m.group(2)}", text)


def normalize_years(text: str, options: NormalizeOptions | None = None) -> str:
    active = options or NormalizeOptions()
    if not active.enable_year_normalization:
        return text

    pattern_range_decade = re.compile(
        r"(?:(?P<prep>в|во|к|ко|с|со|до|от|на|за)\s+)?(?P<year1>\d{2,4})\s*(?:[-–—]|,)\s*(?P<year2>\d{2,4})[-–—](?P<suffix>е|х|м|ми)(?:\s+(?P<word>год[а-яё]*\b))?",
        re.IGNORECASE | re.UNICODE,
    )
    pattern_s_po = re.compile(
        rf"(?P<prep>с|со|от)\s+(?P<year1>{YEAR_ANY_NUMBER_PATTERN})\s+(?P<mid>до|по)\s+(?P<year2>{YEAR_ANY_NUMBER_PATTERN})(?!\d)(?:\s+(?P<word>год[а-яё]*\b|{YEAR_PLURAL_ABBREV_REGEX}(?!\w)|г\.?(?!\w)))?(?:\s+(?P<era>{ERA_REGEX}))?",
        re.IGNORECASE | re.UNICODE,
    )
    pattern_range = re.compile(
        rf"(?:(?P<prep>с|со|из|от|в|во)\s+)?(?P<year1>{YEAR_ANY_NUMBER_PATTERN})\s*[-–—]\s*(?P<year2>{YEAR_ANY_NUMBER_PATTERN})\s+(?P<word>год[а-яё]*\b|{YEAR_PLURAL_ABBREV_REGEX}(?!\w)|г\.?(?!\w))(?:\s+(?P<era>{ERA_REGEX}))?",
        re.IGNORECASE | re.UNICODE,
    )
    pattern_multiple_years = re.compile(
        rf"(?:(?P<prep>{YEAR_PREPOSITION_REGEX})\s+)?(?P<years>{YEAR_RANGE_NUMBER_PATTERN}(?:\s*,\s*{YEAR_RANGE_NUMBER_PATTERN})*\s+и\s+{YEAR_RANGE_NUMBER_PATTERN})\s+(?P<word>год[а-яё]*\b|{YEAR_PLURAL_ABBREV_REGEX}(?!\w))",
        re.IGNORECASE | re.UNICODE,
    )
    pattern_suffix = re.compile(
        rf"(?:^|(?<=\s)|(?<=[«\"'(]))(\d+)[-–—]({YEAR_SUFFIX_REGEX})(?![а-яА-ЯёЁ\d])(?:\s+(год[а-яё]*\b|{YEAR_PLURAL_ABBREV_REGEX}(?!\w)|г\.?(?!\w)))?",
        re.IGNORECASE | re.UNICODE,
    )
    pattern_parenthesized_year = re.compile(
        r"(?P<open>[\(\[])(?P<year>1\d{3}|20\d{2})(?P<close>[\)\]])",
        re.IGNORECASE | re.UNICODE,
    )
    pattern_era_year = re.compile(
        rf"(?:(?P<prep>{YEAR_PREPOSITION_REGEX})\s+)?(?P<year>{YEAR_ANY_NUMBER_PATTERN})(?:\s*[-–—](?P<suffix>{YEAR_SUFFIX_REGEX}))?(?:\s+(?P<word>год[а-яё]*\b|г\.?(?!\w)))?\s+(?P<era>{ERA_REGEX})",
        re.IGNORECASE | re.UNICODE,
    )
    pattern_era_range_shared = re.compile(
        rf"(?:(?P<prep>{YEAR_PREPOSITION_REGEX})\s+)?(?P<year1>{YEAR_ANY_NUMBER_PATTERN})\s*[-–—]\s*(?P<year2>{YEAR_ANY_NUMBER_PATTERN})(?:\s+(?P<word>год[а-яё]*\b|г\.?(?!\w)))?\s+(?P<era>{ERA_REGEX})",
        re.IGNORECASE | re.UNICODE,
    )
    pattern_year_word = re.compile(
        rf"(?:(?P<prep>{YEAR_PREPOSITION_REGEX})\s+)?(?P<year>{YEAR_ANY_NUMBER_PATTERN})\s+(?P<word>год[а-яё]*\b|{YEAR_PLURAL_ABBREV_REGEX}(?!\w)|г\.?(?!\w))",
        re.IGNORECASE | re.UNICODE,
    )
    explicit_year_word_tail_pattern = re.compile(
        rf"\s*(?:год[а-яё]*\b|{YEAR_PLURAL_ABBREV_REGEX}(?!\w)|г\.?(?!\w))",
        re.IGNORECASE | re.UNICODE,
    )
    year_suffix_tail_pattern = re.compile(
        rf"\s*[-–—](?:{YEAR_SUFFIX_REGEX})\b",
        re.IGNORECASE | re.UNICODE,
    )
    pattern_ot_do_implicit = re.compile(
        rf"(?P<prep>с|со|от)\s+(?P<year1>{YEAR_RANGE_NUMBER_PATTERN})\s+(?P<mid>до|по)\s+(?P<year2>{YEAR_RANGE_NUMBER_PATTERN})(?!\d)",
        re.IGNORECASE | re.UNICODE,
    )
    pattern_prep_year_implicit = re.compile(
        rf"\b(?P<prep>{YEAR_IMPLICIT_PREP_REGEX})\s+(?P<year>{YEAR_IMPLICIT_PREP_PATTERN})(?!\d)",
        re.IGNORECASE | re.UNICODE,
    )

    def infer_suffix_case(match: re.Match[str], suffix: str) -> str:
        default_case = YEAR_SUFFIX_TO_CASE.get(suffix, "nomn")
        if suffix not in {"х", "м"}:
            return default_case
        left_context = text[max(0, match.start() - 30) : match.start()]
        prep_match = re.search(
            r"(?:^|[\s«\"'(])(?P<prep>в|во|о|об|к|ко|с|со|до|от|из|по)\s*$",
            left_context,
            re.IGNORECASE | re.UNICODE,
        )
        if not prep_match:
            return default_case
        prep = prep_match.group("prep").lower()
        if suffix == "х":
            if prep in {"в", "во", "о", "об"}:
                return "loct"
            if prep in {"с", "со", "до", "от", "из"}:
                return "gent"
        if suffix == "м":
            if prep in {"в", "во", "о", "об"}:
                return "loct"
            if prep in {"к", "ко", "по"}:
                return "datv"
        return default_case

    def replace_range_decade(m: re.Match[str]) -> str:
        case = YEAR_SUFFIX_TO_CASE.get(m.group("suffix").lower(), "nomn")
        result = (
            (f"{m.group('prep')} " if m.group("prep") else "")
            + f"{year_to_ordinal_words(int(m.group('year1')), case, True)} — {year_to_ordinal_words(int(m.group('year2')), case, True)}"
        )
        if m.group("word"):
            result += f" {m.group('word')}"
        return result

    def replace_s_po(m: re.Match[str]) -> str:
        prep = m.group("prep")
        mid = m.group("mid")
        year1 = int(m.group("year1"))
        year2 = int(m.group("year2"))
        era_text = (
            _normalize_era_marker(m.group("era")) if m.group("era") is not None else None
        )
        if m.group("era") is not None and era_text is None:
            return m.group(0)
        if (
            era_text is None
            and not m.group("word")
            and not (is_plausible_year(year1) and is_plausible_year(year2))
        ):
            return m.group(0)
        if era_text is None and not m.group("word") and not should_treat_as_implicit_year(
            text,
            m.end(),
            explicit_year_word_pattern=explicit_year_word_tail_pattern,
            year_suffix_tail_pattern=year_suffix_tail_pattern,
        ):
            return m.group(0)
        case2 = "accs" if mid.lower() == "по" else "gent"
        result = (
            f"{prep} {year_to_ordinal_words(year1, 'gent')} "
            f"{mid} {year_to_ordinal_words(year2, case2)}"
        )
        word = m.group("word")
        if word:
            word_lower = word.lower()
            if word_lower in ("г.", "г"):
                word_norm = "год"
                word_inflected = YEAR_WORD_FORMS.get((word_norm, case2), word_norm)
            elif re.fullmatch(YEAR_PLURAL_ABBREV_REGEX, word_lower):
                word_norm = "годы"
                word_inflected = YEAR_WORD_FORMS.get((word_norm, case2), word_norm)
            else:
                word_inflected = word
            tail = text[m.end() :].lstrip()
            keep_terminal_dot = not tail or tail[:1].isupper()
            if (
                keep_terminal_dot
                and word.endswith(".")
                and not word_inflected.endswith(".")
            ):
                word_inflected += "."
            result += f" {word_inflected}"
        elif era_text is not None:
            result += f" {YEAR_WORD_FORMS.get(('год', case2), 'год')}"
        if era_text is not None:
            result += f" {era_text}"
            if _should_keep_era_terminal_dot(text, m):
                result += "."
        return result

    def replace_suffix(m: re.Match[str]) -> str:
        year = int(m.group(1))
        suffix = m.group(2).lower()
        word = m.group(3)
        plural = suffix in PLURAL_SUFFIXES
        case = infer_suffix_case(m, suffix)
        if plural:
            if year < 100 and not word:
                return m.group(0)
            # Bare decade-like forms such as "1980-х" are common for years,
            # but model codes like "6301-Х" should not be routed through
            # year normalization when there is no explicit year word.
            if not word and not is_plausible_year(year):
                return m.group(0)
            year = (year // 10) * 10
        elif not word and not is_plausible_year(year):
            return m.group(0)
        result = year_to_ordinal_words(year, case, plural)
        if word:
            word_lower = word.lower()
            if word_lower == "г.":
                word_norm = "год"
                inflected = YEAR_WORD_FORMS.get((word_norm, case), word_norm)
            elif re.fullmatch(YEAR_PLURAL_ABBREV_REGEX, word_lower):
                word_norm = "годы"
                inflected = YEAR_WORD_FORMS.get((word_norm, case), word_norm)
            else:
                inflected = word
            result += f" {inflected}"
        return result

    def replace_multiple_years(m: re.Match[str]) -> str:
        prep = m.group("prep")
        years_text = m.group("years")
        word = m.group("word")
        word_lower = word.lower()
        if re.fullmatch(YEAR_PLURAL_ABBREV_REGEX, word_lower):
            case = PREPOSITIONS_TO_CASE.get(prep.lower(), "nomn") if prep else "nomn"
            word_norm = "годы"
        else:
            case = YEAR_WORD_TO_CASE.get(word_lower, "nomn")
            if prep and case == "nomn":
                case = PREPOSITIONS_TO_CASE.get(prep.lower(), case)
            word_norm = "годы" if word_lower.startswith("год") else word

        parts = re.split(r"(\s*,\s*|\s+и\s+)", years_text)
        rendered_parts: list[str] = []
        for part in parts:
            stripped = part.strip()
            if not stripped:
                continue
            if stripped.isdigit():
                rendered_parts.append(year_to_ordinal_words(int(stripped), case))
            else:
                rendered_parts.append(part)

        prefix = f"{prep} " if prep else ""
        inflected_word = YEAR_WORD_FORMS.get((word_norm, case), word_norm)
        return f"{prefix}{''.join(rendered_parts)} {inflected_word}"

    def replace_era_year(m: re.Match[str]) -> str:
        prep = m.group("prep")
        word = m.group("word")
        suffix = m.group("suffix")
        era_text = _normalize_era_marker(m.group("era"))
        if era_text is None:
            return m.group(0)
        if suffix:
            case = YEAR_SUFFIX_TO_CASE.get(suffix.lower(), "nomn")
            if suffix.lower() == "м":
                case = "loct"
        elif word:
            word_lower = word.lower()
            if word_lower in ("г.", "г"):
                case = PREPOSITIONS_TO_CASE.get(prep.lower(), "nomn") if prep else "nomn"
            else:
                case = _resolve_explicit_year_word_case(prep, word)
        elif prep:
            case = PREPOSITIONS_TO_CASE.get(prep.lower(), "nomn")
        else:
            case = "nomn"
        if word and word.lower() in ("г.", "г"):
            word_form = YEAR_WORD_FORMS.get(("год", case), "год")
        elif word:
            word_form = word
        else:
            word_form = YEAR_WORD_FORMS.get(("год", case), "год")
        prefix = f"{prep} " if prep else ""
        result = (
            f"{prefix}{year_to_ordinal_words(int(m.group('year')), case)} "
            f"{word_form} {era_text}"
        )
        if _should_keep_era_terminal_dot(text, m):
            result += "."
        return result

    def replace_era_range_shared(m: re.Match[str]) -> str:
        prep = m.group("prep")
        word = m.group("word")
        era_text = _normalize_era_marker(m.group("era"))
        if era_text is None:
            return m.group(0)
        if word:
            word_lower = word.lower()
            if word_lower in ("г.", "г"):
                case = PREPOSITIONS_TO_CASE.get(prep.lower(), "nomn") if prep else "nomn"
            else:
                case = _resolve_explicit_year_word_case(prep, word)
        elif prep:
            case = PREPOSITIONS_TO_CASE.get(prep.lower(), "nomn")
        else:
            case = "nomn"

        if word and word.lower() in ("г.", "г"):
            word_form = YEAR_WORD_FORMS.get(("год", case), "год")
        elif word:
            word_form = word
        else:
            word_form = YEAR_WORD_FORMS.get(("год", case), "год")

        prefix = f"{prep} " if prep else ""
        result = (
            f"{prefix}{year_to_ordinal_words(int(m.group('year1')), case)} — "
            f"{year_to_ordinal_words(int(m.group('year2')), case)} "
            f"{word_form} {era_text}"
        )
        if _should_keep_era_terminal_dot(text, m):
            result += "."
        return result

    def replace_parenthesized_year(m: re.Match[str]) -> str:
        year = int(m.group("year"))
        if not is_plausible_year(year):
            return m.group(0)
        left_context = text[max(0, m.start() - 40) : m.start()]
        left_tokens = [
            normalize_context_token(token)
            for token in simple_tokenize(left_context)
            if normalize_context_token(token)
        ]
        if len(left_tokens) >= 2 and left_tokens[-2] in {"под", "при"}:
            return f"{m.group('open')}{year_to_ordinal_words(year)}{m.group('close')}"
        try:
            rendered = num2words.num2words(year, lang="ru")
        except Exception:
            rendered = str(year)
        return f"{m.group('open')}{rendered}{m.group('close')}"

    def replace_with_word(m: re.Match[str]) -> str:
        prep = m.group("prep")
        year = int(m.group("year"))
        word = m.group("word")
        word_lower = word.lower()
        if word_lower == "г.":
            word_norm = "год"
            is_abbrev = True
            plural = False
        elif re.fullmatch(YEAR_PLURAL_ABBREV_REGEX, word_lower):
            word_norm = "годы"
            is_abbrev = True
            plural = True
        else:
            word_norm = word
            is_abbrev = False
            plural = word_lower in ("годы", "годов", "годам", "годами", "годах")

        is_year_prepositional = word_lower == "году" and prep
        if not is_year_prepositional and year < 1000 and not is_abbrev:
            return m.group(0)
        if not is_abbrev:
            case = _resolve_explicit_year_word_case(prep, word)
        elif prep:
            case = PREPOSITIONS_TO_CASE.get(prep.lower(), "nomn")
        else:
            case = "gent"

        ordinal = year_to_ordinal_words(year, case, plural)
        prefix = f"{prep} " if prep else ""
        if is_abbrev:
            return (
                f"{prefix}{ordinal} {YEAR_WORD_FORMS.get((word_norm, case), word_norm)}"
            )
        return f"{prefix}{ordinal} {word}"

    def replace_range(m: re.Match[str]) -> str:
        prep = m.group("prep")
        word = m.group("word")
        era_text = (
            _normalize_era_marker(m.group("era")) if m.group("era") is not None else None
        )
        if m.group("era") is not None and era_text is None:
            return m.group(0)
        word_lower = word.lower()
        if word_lower == "г." or re.fullmatch(YEAR_PLURAL_ABBREV_REGEX, word_lower):
            word_norm = "годы"
            is_abbrev = True
        else:
            word_norm = word
            is_abbrev = False
        if word and not is_abbrev:
            case = YEAR_WORD_TO_CASE.get(word_lower, "nomn")
        elif prep:
            case = PREPOSITIONS_TO_CASE.get(prep.lower(), "nomn")
        else:
            case = "nomn"
        result = (
            (f"{prep} " if prep else "")
            + f"{year_to_ordinal_words(int(m.group('year1')), case, True)} — {year_to_ordinal_words(int(m.group('year2')), case, True)}"
        )
        if word_norm:
            if is_abbrev:
                inflected = YEAR_WORD_FORMS.get((word_norm, case), word_norm)
                if word.endswith(".") and not inflected.endswith("."):
                    inflected += "."
                result += f" {inflected}"
            else:
                result += f" {word}"
        if era_text is not None:
            result += f" {era_text}"
            if _should_keep_era_terminal_dot(text, m):
                result += "."
        return result

    def replace_ot_do_implicit(m: re.Match[str]) -> str:
        year1 = int(m.group("year1"))
        year2 = int(m.group("year2"))
        if not (is_plausible_year(year1) and is_plausible_year(year2)):
            return m.group(0)
        if not should_treat_as_implicit_year(
            text,
            m.end(),
            explicit_year_word_pattern=explicit_year_word_tail_pattern,
            year_suffix_tail_pattern=year_suffix_tail_pattern,
        ):
            return m.group(0)
        case2 = "accs" if m.group("mid").lower() == "по" else "gent"
        return f"{m.group('prep')} {year_to_ordinal_words(year1, 'gent')} {m.group('mid')} {year_to_ordinal_words(year2, case2)}"

    def replace_prep_year_implicit(m: re.Match[str]) -> str:
        if not should_treat_as_implicit_year(
            text,
            m.end(),
            explicit_year_word_pattern=explicit_year_word_tail_pattern,
            year_suffix_tail_pattern=year_suffix_tail_pattern,
        ):
            return m.group(0)
        prep = m.group("prep").lower()
        if prep in ("в", "во", "о", "об"):
            case = "loct"
        elif prep in ("к", "ко", "по"):
            case = "datv"
        elif prep in ("с", "со", "до", "от", "из"):
            case = "gent"
        elif prep in ("на", "за", "под"):
            case = "accs"
        else:
            case = "nomn"
        return f"{m.group('prep')} {year_to_ordinal_words(int(m.group('year')), case)}"

    text = pattern_range_decade.sub(replace_range_decade, text)
    text = pattern_s_po.sub(replace_s_po, text)
    text = pattern_range.sub(replace_range, text)
    text = pattern_multiple_years.sub(replace_multiple_years, text)
    text = pattern_parenthesized_year.sub(replace_parenthesized_year, text)
    text = pattern_era_range_shared.sub(replace_era_range_shared, text)
    text = pattern_era_year.sub(replace_era_year, text)
    text = pattern_suffix.sub(replace_suffix, text)
    text = pattern_ot_do_implicit.sub(replace_ot_do_implicit, text)
    text = pattern_prep_year_implicit.sub(replace_prep_year_implicit, text)
    text = pattern_year_word.sub(replace_with_word, text)
    return re.sub(
        rf"(?<![А-Яа-яA-Za-z_]){YEAR_PLURAL_ABBREV_REGEX}(?!\w)",
        "годы",
        text,
        flags=re.IGNORECASE,
    )



from __future__ import annotations

import re

from ._morph import get_morph
from .constants import KNOWN_ABBREVIATIONS
from .preprocess_utils import PARAGRAPH_BREAK_PLACEHOLDER

_LINE_SPLIT_PATTERN = re.compile(rf"(\n|{re.escape(PARAGRAPH_BREAK_PLACEHOLDER)})")
_SENTENCE_START_PATTERN = re.compile(
    r"((?<=[.!?…])(?:[\s\"'«»„“”()\[\]{}]*))([a-zA-Zа-яА-ЯёЁ])"
)


def normalize_first_word_caps(text: str, enabled: bool = True) -> str:
    if not enabled:
        return text

    parts = _LINE_SPLIT_PATTERN.split(text)
    first_word_pattern = re.compile(r"^(\s*)([A-ZА-ЯЁ]{2,})(?=\b|\s|[.,!?;:])")

    for idx in range(0, len(parts), 2):
        line = parts[idx]
        match = first_word_pattern.match(line)
        if match:
            word = match.group(2)
            rest = line[match.end() :]
            has_following_letter_or_digit = any(
                char.isalpha() or char.isdigit() for char in rest
            )
            if (
                word not in KNOWN_ABBREVIATIONS
                and len(word) > 3
                and has_following_letter_or_digit
            ):
                parts[idx] = match.group(1) + word.capitalize() + line[match.end() :]

    return "".join(parts)


def normalize_sentence_start_caps(text: str, enabled: bool = True) -> str:
    if not enabled:
        return text

    def uppercase_sentence_start(match: re.Match[str]) -> str:
        prefix = match.group(1)
        letter = match.group(2)
        return f"{prefix}{letter.upper()}"

    parts = _LINE_SPLIT_PATTERN.split(text)
    for idx in range(0, len(parts), 2):
        parts[idx] = _SENTENCE_START_PATTERN.sub(uppercase_sentence_start, parts[idx])
    return "".join(parts)


def _is_caps_token(token: str) -> bool:
    """Check if a whitespace-split token is an all-caps non-abbreviation word."""
    letters_only = re.sub(r"[^a-zA-Zа-яА-ЯёЁ]", "", token)
    return (
        len(letters_only) >= 2
        and letters_only.isupper()
        and letters_only not in KNOWN_ABBREVIATIONS
    )


def _normalize_inline_caps(line: str) -> str:
    tokens = re.split(r"(\s+)", line)

    i = 0
    while i < len(tokens):
        token = tokens[i]
        if token.isspace():
            i += 1
            continue

        if _is_caps_token(token):
            run_start = i
            caps_count = 1
            j = i + 1
            while j < len(tokens):
                if tokens[j].isspace():
                    j += 1
                    continue
                if _is_caps_token(tokens[j]):
                    caps_count += 1
                    j += 1
                else:
                    break
            if caps_count >= 2:
                first_word = True
                for k in range(run_start, j):
                    if not tokens[k].isspace():
                        if first_word:
                            tokens[k] = _lowercase_preserve_abbrevs(
                                tokens[k]
                            ).capitalize()
                            first_word = False
                        else:
                            tokens[k] = _lowercase_preserve_abbrevs(tokens[k])
            i = j
        else:
            i += 1
    return "".join(tokens)


def _lowercase_preserve_abbrevs(token: str) -> str:
    """Lowercase a token, but preserve known abbreviation substrings."""
    letters = re.sub(r"[^a-zA-Zа-яА-ЯёЁ]", "", token)
    if letters.isupper() and letters in KNOWN_ABBREVIATIONS:
        return token
    return token.lower()


def _restore_known_abbreviations(original: str, lowered: str) -> str:
    """After lowering, restore tokens that were originally known abbreviations."""
    orig_tokens = re.split(r"(\s+)", original)
    low_tokens = re.split(r"(\s+)", lowered)
    if len(orig_tokens) != len(low_tokens):
        return lowered
    vowels = set("АЕЁИОУЫЭЮЯ")
    for i, orig_tok in enumerate(orig_tokens):
        letters = re.sub(r"[^a-zA-Zа-яА-ЯёЁ]", "", orig_tok)
        if not letters or not letters.isupper():
            continue
        upper_letters = letters.upper()
        is_known = upper_letters in KNOWN_ABBREVIATIONS
        is_short_abbrev = len(letters) <= 3
        is_real_short_word = False
        if 2 <= len(letters) <= 3:
            parsed = get_morph().parse(letters.lower())[0]
            is_real_short_word = parsed.tag.POS is not None
        is_consonant_heavy = (
            len(letters) == 4 and sum(1 for c in upper_letters if c in vowels) <= 1
        )
        if is_known or ((is_short_abbrev and not is_real_short_word)) or is_consonant_heavy:
            low_tokens[i] = orig_tok
    return "".join(low_tokens)


def normalize_caps_lines(
    text: str, enabled: bool = True, caps_threshold: float = 0.8, min_length: int = 10
) -> str:
    if not enabled:
        return text

    parts = _LINE_SPLIT_PATTERN.split(text)
    for idx in range(0, len(parts), 2):
        line = parts[idx]
        letters = [char for char in line if char.isalpha()]
        if len(letters) >= min_length:
            caps_ratio = sum(1 for char in letters if char.isupper()) / len(letters)
            if caps_ratio >= caps_threshold:
                stripped = line.strip()
                lowered = stripped.lower()
                lowered = _restore_known_abbreviations(stripped, lowered)
                for ci, ch in enumerate(lowered):
                    if ch.isalpha():
                        lowered = lowered[:ci] + ch.upper() + lowered[ci + 1 :]
                        break
                parts[idx] = (
                    line[: len(line) - len(line.lstrip())]
                    + lowered
                    + line[len(line.rstrip()) :]
                )
                continue
        parts[idx] = _normalize_inline_caps(line)
    return "".join(parts)



from __future__ import annotations

import argparse
import sys
from pathlib import Path

from .options import NormalizeOptions
from .pipeline import normalize


def _read_input(args: argparse.Namespace) -> str:
    if args.file:
        return Path(args.file).read_text(encoding="utf-8")
    if args.text:
        return args.text
    if not sys.stdin.isatty():
        return sys.stdin.read()
    raise SystemExit("Provide text, --file, or stdin.")


def build_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog="ru-normalizr", description="Normalize Russian text."
    )
    parser.add_argument("text", nargs="?", help="Inline text to normalize.")
    parser.add_argument("--file", help="Read text from file.")
    parser.add_argument(
        "--output", help="Write normalized text to file instead of stdout."
    )
    parser.add_argument(
        "--check", action="store_true", help="Normalize input and print the result."
    )
    parser.add_argument(
        "--mode",
        choices=["safe", "tts"],
        default="safe",
        help="Preset normalization mode.",
    )
    parser.add_argument(
        "--latinization-backend",
        choices=["ipa", "dictionary"],
        help="Backend for Latin transliteration.",
    )
    parser.add_argument(
        "--no-latinization", action="store_true", help="Disable Latin transliteration."
    )
    parser.add_argument(
        "--no-first-word-decap", action="store_true", help="Disable first-word decap."
    )
    parser.add_argument(
        "--keep-links", action="store_true", help="Keep bracketed numeric links."
    )
    parser.add_argument(
        "--with-latin-stress",
        action="store_true",
        help="Keep '+' stress markers when using IPA latinization.",
    )
    return parser


def main(argv: list[str] | None = None) -> int:
    parser = build_parser()
    args = parser.parse_args(argv)
    text = _read_input(args)
    options = NormalizeOptions(
        mode=args.mode,
        enable_first_word_decap=False if args.no_first_word_decap else None,
        remove_links=False if args.keep_links else None,
        enable_latinization=False if args.no_latinization else None,
        latinization_backend=args.latinization_backend,
        enable_latinization_stress_marks=args.with_latin_stress,
    )
    result = normalize(text, options)
    if args.output:
        output_path = Path(args.output)
        output_text = result if result.endswith("\n") else result + "\n"
        output_path.write_text(output_text, encoding="utf-8")
    else:
        sys.stdout.write(result)
        if not result.endswith("\n"):
            sys.stdout.write("\n")
    return 0


if __name__ == "__main__":
    raise SystemExit(main())



from __future__ import annotations

import re

from .._morph import get_morph
from ..preprocess_utils import NEGATIVE_NUMBER_PLACEHOLDER
from ..text_context import simple_tokenize
from ._constants import PREP_CASE, UNIT_TOKEN_FRAGMENT, UNITS_DATA
from ._helpers import (
    get_numeral_case,
    inflect_numeral_string,
    inflect_unit_lemma,
    safe_inflect,
    should_keep_decimal_unit_dot,
)

DECIMAL_PATTERN = re.compile(
    rf"(?<!\d)(?P<num>(?:-|{re.escape(NEGATIVE_NUMBER_PLACEHOLDER)})?\d+[.,]\d+)(?:\s*(?P<unit>{UNIT_TOKEN_FRAGMENT})(?P<unit_dot>\.)?)?(?:\s+(?P<unit2>{UNIT_TOKEN_FRAGMENT})(?P<unit2_dot>\.)?)?(?!\d)"
)


def normalize_decimals(text: str) -> str:
    def is_ambiguous_preposition_token(token: str) -> bool:
        clean = token.lower().strip('.,:;!"«»()[]{}')
        return bool(clean) and " " not in clean and clean in PREP_CASE

    def should_skip_unit_candidate(unit_raw: str, rest: str) -> bool:
        if not is_ambiguous_preposition_token(unit_raw):
            return False
        stripped_rest = rest.lstrip()
        if not stripped_rest:
            return False
        next_char = stripped_rest[:1]
        return bool(re.match(r"[^\W_]", next_char, re.UNICODE))

    def should_skip_combined_unit_candidate(unit_parts: list[str]) -> bool:
        cleaned_parts = [part.strip('.,:;!"«»()[]{}') for part in unit_parts]
        return any(
            cleaned and is_ambiguous_preposition_token(cleaned)
            for cleaned in cleaned_parts[1:]
        )

    def repl(match: re.Match[str]) -> str:
        s = match.group("num").replace(",", ".")
        unit_raw = match.group("unit")
        start_pos = match.start()
        context = text[max(0, start_pos - 40) : start_pos]
        tokens_left = simple_tokenize(context)
        case = get_numeral_case(tokens_left + [s], len(tokens_left))
        is_negative = s.startswith("-") or s.startswith(NEGATIVE_NUMBER_PLACEHOLDER)
        abs_s = s.lstrip(f"-{NEGATIVE_NUMBER_PLACEHOLDER}")
        parts = abs_s.split(".")
        if len(parts) != 2:
            return match.group(0)
        int_part_s, frac_part_s = parts
        int_val = int(int_part_s)
        frac_val = int(frac_part_s)
        digits = len(frac_part_s)
        order_names = {
            1: "десятая",
            2: "сотая",
            3: "тысячная",
            4: "десятитысячная",
            5: "стотысячная",
            6: "миллионная",
        }
        order_name_base = order_names.get(digits, "десятитысячная")
        morph = get_morph()
        int_words = inflect_numeral_string(int_part_s, case, gender="femn")
        p_cel = morph.parse("целая")[0]
        tags_cel = (
            {case, "femn", "sing"}
            if int_val % 10 == 1 and int_val % 100 != 11
            else ({"gent", "plur"} if case in ["nomn", "accs"] else {case, "plur"})
        )
        cel_words = safe_inflect(p_cel, tags_cel)
        frac_words = inflect_numeral_string(frac_part_s, case, gender="femn")
        p_order = morph.parse(order_name_base)[0]
        tags_order = (
            {case, "femn", "sing"}
            if frac_val % 10 == 1 and frac_val % 100 != 11
            else ({"gent", "plur"} if case in ["nomn", "accs"] else {case, "plur"})
        )
        order_words = safe_inflect(p_order, tags_order)
        result = f"{int_words} {cel_words} {frac_words} {order_words}"
        if unit_raw:
            unit_dot = match.group("unit_dot")
            unit2_raw = match.group("unit2")
            unit2_dot = match.group("unit2_dot")
            unit_info = None
            unit_consumes_second_token = False
            if unit2_raw and not should_skip_combined_unit_candidate([unit_raw, unit2_raw]):
                combined_candidates = []
                if unit_dot:
                    combined_candidates.append(f"{unit_raw}.{unit2_raw}")
                combined_candidates.append(f"{unit_raw}{unit2_raw}")
                for candidate_raw in combined_candidates:
                    candidate_key = candidate_raw.lower().strip(".")
                    unit_info = UNITS_DATA.get(candidate_key)
                    if unit_info:
                        unit_consumes_second_token = True
                        break
            if unit_info is None:
                unit_lower = unit_raw.lower().strip(".")
                if not should_skip_unit_candidate(unit_raw, text[match.end("unit") :]):
                    unit_info = UNITS_DATA.get(unit_lower)
            unit2_processed = False
            if unit_info:
                lemma, _, _, *u_suffix = unit_info
                result += " " + inflect_unit_lemma(lemma, {"gent", "sing"})
                if u_suffix:
                    result += " " + u_suffix[0]
                if unit_consumes_second_token:
                    unit2_processed = True
                elif unit2_raw:
                    unit2_lower = unit2_raw.lower().strip(".")
                    unit2_info = UNITS_DATA.get(unit2_lower)
                    multipliers = {"тысяча", "миллион", "миллиард", "триллион"}
                    if lemma in multipliers and unit2_info:
                        lemma2, _, _, *suffix2 = unit2_info
                        result += " " + inflect_unit_lemma(lemma2, {"gent", "plur"})
                        if suffix2:
                            result += " " + suffix2[0]
                        unit2_processed = True
                if unit2_raw and not unit2_processed:
                    result += " " + unit2_raw
            else:
                result += " " + unit_raw
                if unit2_raw:
                    result += " " + unit2_raw
            if (
                unit2_dot
                and unit2_raw
                and should_keep_decimal_unit_dot(text[match.end() :])
            ):
                result += "."
            elif (
                unit_dot
                and not unit2_raw
                and should_keep_decimal_unit_dot(text[match.end() :])
            ):
                result += "."
        if is_negative:
            result = "минус " + result
        return result

    return DECIMAL_PATTERN.sub(repl, text)



from __future__ import annotations

import re
from dataclasses import dataclass


@dataclass(frozen=True, slots=True)
class RegexRule:
    pattern: str
    flags: int = re.IGNORECASE


def rx(pattern: str, flags: int = re.IGNORECASE) -> RegexRule:
    return RegexRule(pattern=pattern, flags=flags)


ABBREVIATION_RULES = {
    rx(r"\bсм\.\s*рис\.(?!\w)"): "смотри рисунок",
    rx(r"\bсм\."): "смотри",
    rx(r"\bнапр\."): "например",
    rx(r"\bт\.?\s*е\."): "то есть",
    rx(r"\bт\.?\s*к\."): "так как",
    rx(r"(?<!\w)н\.?\s*э\.(?!\w)"): "нашей эры",
    rx(r"(?<!\w)нэ\.?(?!\w)"): "нашей эры",
    rx(r"\bс\.\s*(?=\d)"): "страница ",
    rx(r"\bp\.\s*(?=\d)"): "страница ",
    rx(r"\bст\.\s*(?=\d)"): "статья ",
    rx(r"\bгл\."): "глава",
    rx(r"\bmr\."): "мистер",
    rx(r"\bmrs\."): "миссис",
    rx(r"\bг-?н\.?\b"): "господин",
    rx(r"\bг-?жа\.?\b"): "госпожа",
    rx(r"\bСв\.\s+(?=[А-ЯЁ][а-яё]+(?:а|я)\b)", 0): "Святого ",
    rx(r"\bсв\.\s+(?=[А-ЯЁ][а-яё]+(?:а|я)\b)"): "святого ",
    rx(r"\bрис\.\s*(?=\d)"): "рисунок ",
    # rx(r"\bим\."): "имени",
    rx(r"\bул\.\s*(?=[А-ЯЁ])"): "улица ",
    "снск.": "сноска",
    "прим. пер.": "примечание переводчика",
    "прим.пер.": "примечание переводчика",
    "прим. перев.": "примечание переводчика",
    "прим.перев.": "примечание переводчика",
    "прим. ред.": "примечание редактора",
    "прим.ред.": "примечание редактора",
    rx(r"\bперев\."): "переводчик",
    "изд-во": "издательство",
    "ин-т": "институт",
    "инф-ция": "информация",
    "орг-ция": "организация",
    "мин-во": "министерство",
    "гос-во": "государство",
    "гос.": "государственный",
    "междунар.": "международный",
    "обл.": "область",
    "р-н": "район",
    "пос.": "посёлок",
    "пгт": "посёлок городского типа",
    "оз.": "озеро",
    "полит.": "политический",
    "офиц.": "официальный",
    "юридич.": "юридический",
    "ж.-д.": "железнодорожный",
    "с.-х.": "сельскохозяйственный",
    "штаб-кв.": "штаб-квартира",
    "т.д.": "так далее",
    "т. д.": "так далее",
    "т.п.": "тому подобное",
    "т. п.": "тому подобное",
    "табл.": "таблица",
    "стр.": "страница",
    "ред.": "редактор",
    "сост.": "составитель",
    "прил.": "приложение",
    "прим. авт.": "примечание автора",
    "прим.": "примечание",
    "УК РФ": "уголовный кодекс российской федерации",
    "АК РФ": "арбитражный кодекс российской федерации",
    "псевд.": "псевдоним",
    "респ.": "республика",
}

ADJECTIVE_ABBREVIATION_EXPANSIONS = {
    "англ.": "английский",
    "руск.": "русский",
    "рус.": "русский",
    "немецк.": "немецкий",
    "нем.": "немецкий",
    "франц.": "французский",
    "греч.": "греческий",
    "латин.": "латинский",
    "лат.": "латинский",
    "гос.": "государственный",
    "междунар.": "международный",
    "полит.": "политический",
    "офиц.": "официальный",
    "юридич.": "юридический",
    "ж.-д.": "железнодорожный",
    "с.-х.": "сельскохозяйственный",
}


def _compile_literal_rule(abbr: str) -> re.Pattern[str]:
    return re.compile(rf"(?<!\w){re.escape(abbr)}(?!\w)", re.IGNORECASE)


ABBREVIATION_PATTERNS = [
    (
        (
            re.compile(matcher.pattern, matcher.flags)
            if isinstance(matcher, RegexRule)
            else _compile_literal_rule(matcher)
        ),
        replacement,
    )
    for matcher, replacement in ABBREVIATION_RULES.items()
]



from __future__ import annotations

import re

import num2words

from .._morph import get_morph
from ..ordinal_utils import render_ordinal
from ..text_context import simple_tokenize
from ._constants import HYPHENATED_WORD_PATTERN, ORDINAL_PATTERN
from ._helpers import get_numeral_case, inflect_numeral_string
from ._hyphen import CARDINAL_CASE_SUFFIXES, classify_numeric_hyphen_rhs

HEADING_WORDS_PATTERN = (
    r"глава|главы|главе|главу|главой|главами|главах|"
    r"часть|части|частью|частях|"
    r"раздел|раздела|разделе|разделу|разделом|разделах|"
    r"том|тома|томе|томом|томах|"
    r"книга|книги|книге|книгой|книгах|"
    r"квартал|квартала|квартале|кварталу|кварталом|кварталах"
)
AMBIGUOUS_HEADING_WORDS = {
    "главы",
    "части",
    "раздела",
    "тома",
    "книги",
    "квартала",
}
HEADING_CONTEXT_CASES = {
    "в": "loct",
    "во": "loct",
    "о": "loct",
    "об": "loct",
    "обо": "loct",
    "при": "loct",
    "к": "datv",
    "ко": "datv",
    "по": "datv",
    "с": "gent",
    "со": "gent",
    "из": "gent",
    "до": "gent",
    "от": "gent",
    "у": "gent",
    "без": "gent",
    "после": "gent",
    "для": "gent",
    "около": "gent",
    "возле": "gent",
    "вокруг": "gent",
    "кроме": "gent",
    "начало": "gent",
    "конец": "gent",
    "середина": "gent",
}
SINGULAR_HEADING_WORDS_PATTERN = (
    r"глава|главы|главе|главу|главой|"
    r"часть|части|частью|"
    r"раздел|раздела|разделе|разделу|"
    r"том|тома|томе|томом|"
    r"книга|книги|книге|книгой|"
    r"квартал|квартала|квартале|кварталу|кварталом"
)
HEADING_RANGE_PATTERN = re.compile(
    rf"\b(?P<head>{HEADING_WORDS_PATTERN})\s+(?P<left>\d+)\s*[–—-]\s*(?P<right>\d+)\b",
    re.IGNORECASE | re.UNICODE,
)
HEADING_SINGLE_PATTERN = re.compile(
    rf"\b(?P<head>{SINGULAR_HEADING_WORDS_PATTERN})\s+(?P<number>\d+)\b",
    re.IGNORECASE | re.UNICODE,
)


def _ordinal_words(num: int, case: str, gender: str | None) -> str:
    return render_ordinal(num, case=case, gender=gender)


def _pick_range_preposition(first_ordinal: str) -> str:
    return "со" if first_ordinal.startswith(("в", "ф", "с", "з", "ш", "ж")) else "с"


def _heading_parse(text: str, match_start: int, head: str):
    parsed = [candidate for candidate in get_morph().parse(head.lower()) if "NOUN" in candidate.tag]
    if not parsed:
        return None
    inanimate = [candidate for candidate in parsed if "inan" in candidate.tag]
    candidates = inanimate if inanimate else parsed
    normalized_head = head.lower()
    if normalized_head not in AMBIGUOUS_HEADING_WORDS:
        nominative_singular = [
            candidate
            for candidate in candidates
            if ("sing" in candidate.tag and "nomn" in candidate.tag)
        ]
        if nominative_singular:
            return nominative_singular[0]
        return candidates[0]

    left_context = text[max(0, match_start - 40) : match_start]
    left_tokens = simple_tokenize(left_context)
    left_word = next(
        (
            token.lower()
            for token in reversed(left_tokens)
            if any(char.isalpha() for char in token)
        ),
        "",
    )
    target_case = HEADING_CONTEXT_CASES.get(left_word)
    if target_case is None:
        return None

    for candidate in candidates:
        candidate_case = "loct" if "loc2" in candidate.tag else candidate.tag.case
        if candidate_case == target_case and "sing" in candidate.tag:
            return candidate
    for candidate in candidates:
        candidate_case = "loct" if "loc2" in candidate.tag else candidate.tag.case
        if candidate_case == target_case:
            return candidate
    return None


def normalize_heading_ranges(text: str) -> str:
    def repl(match: re.Match[str]) -> str:
        head = match.group("head")
        noun_parse = _heading_parse(text, match.start(), head)
        gender = noun_parse.tag.gender if noun_parse and noun_parse.tag.gender else "masc"
        left_ordinal = _ordinal_words(int(match.group("left")), "gent", gender)
        right_case = "accs" if gender == "femn" else "nomn"
        right_ordinal = _ordinal_words(int(match.group("right")), right_case, gender)
        return f"{head} {_pick_range_preposition(left_ordinal)} {left_ordinal} по {right_ordinal}"

    return HEADING_RANGE_PATTERN.sub(repl, text)


def normalize_heading_numbers(text: str) -> str:
    def repl(match: re.Match[str]) -> str:
        head = match.group("head")
        noun_parse = _heading_parse(text, match.start(), head)
        if noun_parse is None:
            return match.group(0)
        gender = noun_parse.tag.gender or "masc"
        case = noun_parse.tag.case or "nomn"
        ordinal = _ordinal_words(int(match.group("number")), case, gender)
        return f"{head} {ordinal}"

    return HEADING_SINGLE_PATTERN.sub(repl, text)


def normalize_hyphenated_words(text: str) -> str:
    def repl(match: re.Match[str]) -> str:
        morph = get_morph()
        num_str = match.group(1)
        word = match.group(2)
        word_lower = word.lower()
        word_kind = classify_numeric_hyphen_rhs(word)
        if word_kind == "unit":
            return match.group(0)
        if word_kind == "ordinal_suffix":
            return match.group(0)
        if (
            word_lower
            in {
                "ый",
                "ой",
                "й",
                "ого",
                "го",
                "ому",
                "ым",
                "ом",
                "му",
                "е",
                "х",
                "м",
                "ми",
            }
            and int(num_str) > 100
        ):
            return match.group(0)
        try:
            int(num_str)
        except Exception:
            return match.group(0)
        case_from_suffix = None
        if word_lower == "х":
            case_from_suffix = None
        elif word_lower in ("ти", "и"):
            case_from_suffix = "gent"
        elif word_lower in ("ми", "мя"):
            case_from_suffix = "ablt"
        ctx_left = text[max(0, match.start() - 60) : match.start()]
        ctx_right = text[match.end() : match.end() + 60]
        tokens_left = simple_tokenize(ctx_left)
        tokens_right = simple_tokenize(ctx_right)
        context_case = get_numeral_case(
            tokens_left + [num_str] + tokens_right, len(tokens_left)
        )
        if case_from_suffix:
            case = case_from_suffix
        elif word_lower == "х":
            case = context_case if context_case in ("gent", "loct") else "gent"
        else:
            case = context_case
        p_word = morph.parse(word_lower)[0]
        is_adj_like = "ADJF" in p_word.tag or word_lower.endswith(
            (
                "дневный",
                "часовой",
                "минутный",
                "летний",
                "этажный",
                "тонный",
                "процентный",
                "кратный",
                "кратного",
                "кратном",
                "кратных",
            )
        )
        target_case = "gent" if is_adj_like and case == "nomn" else case
        num_words = inflect_numeral_string(num_str, target_case)
        if word_lower in CARDINAL_CASE_SUFFIXES:
            return num_words
        return (
            f"{num_words}{word}"
            if is_adj_like
            else (f"{num_words}" if len(word) <= 3 else f"{num_words} {word}")
        )

    return HYPHENATED_WORD_PATTERN.sub(repl, text)


def normalize_ordinals(text: str) -> str:
    def repl(match: re.Match[str]) -> str:
        morph = get_morph()
        num_str = match.group(1)
        suffix = match.group(2).lower()
        try:
            num = int(num_str)
        except ValueError:
            return match.group(0)
        gender = "masc"
        is_cardinal_suffix = suffix in ("ти", "ми")
        ctx_left = text[max(0, match.start() - 60) : match.start()]
        ctx_right = text[match.end() : match.end() + 60]
        tokens_left = simple_tokenize(ctx_left)
        tokens_right = simple_tokenize(ctx_right)
        case = get_numeral_case(
            tokens_left + [num_str] + tokens_right, len(tokens_left)
        )

        def first_noun_right():
            for token in tokens_right[:4]:
                clean = token.strip(".,!?;:")
                if not clean:
                    continue
                parsed = morph.parse(clean)[0]
                if "NOUN" in parsed.tag:
                    return parsed
                if parsed.tag.POS in {"VERB", "INFN"}:
                    break
            return None

        if suffix in ("я", "яя"):
            gender = "femn"
        elif suffix in ("е", "ее"):
            gender = "neut"
        elif suffix == "й" and tokens_right:
            p_next = morph.parse(tokens_right[0].strip(".,!?;:"))[0]
            if "femn" in p_next.tag:
                gender = "femn"
                if case == "nomn":
                    case = "gent"
        case_from_suffix = None
        if suffix == "го":
            case_from_suffix = "gent"
        elif suffix == "му":
            case_from_suffix = "datv"
        elif suffix == "й" and case == "nomn" and tokens_right:
            p_next = morph.parse(tokens_right[0].strip(".,!?;:"))[0]
            if "femn" in p_next.tag:
                gender = "femn"
                case_from_suffix = "gent"
        if suffix == "м":
            if tokens_right:
                p_next = morph.parse(tokens_right[0].strip(".,!?;:"))[0]
                case_from_suffix = "loct" if "sing" in p_next.tag else "datv"
            else:
                case_from_suffix = "loct"
        case = case_from_suffix or case
        plural = suffix in ("х", "ми", "е", "м") and not (
            suffix == "м" and case == "loct"
        )
        if suffix == "е" and tokens_right:
            next_clean = tokens_right[0].strip(".,!?;:")
            if next_clean:
                p_next = morph.parse(next_clean)[0]
                if "sing" in p_next.tag and "neut" in p_next.tag:
                    plural = False

        generation_case = case
        right_noun = first_noun_right()
        if (
            case == "accs"
            and not plural
            and gender in {"masc", "neut"}
            and right_noun is not None
            and "inan" in right_noun.tag
        ):
            generation_case = "nomn"

        if is_cardinal_suffix:
            cases_map = {
                "nomn": "nominative",
                "gent": "genitive",
                "datv": "dative",
                "accs": "accusative",
                "ablt": "instrumental",
                "loct": "prepositional",
            }
            try:
                return (
                    num2words.num2words(
                        num,
                        lang="ru",
                        case=cases_map.get(generation_case, "nominative"),
                    )
                    + " "
                )
            except Exception:
                return match.group(0)
        return (
            render_ordinal(
                num,
                case=generation_case,
                gender=gender,
                plural=plural,
                inanimate=right_noun is not None and "inan" in right_noun.tag,
            )
            + " "
        )

    return ORDINAL_PATTERN.sub(repl, text)



from __future__ import annotations

import re

from ._morph import get_morph
from .abbreviation_rules import (
    ABBREVIATION_PATTERNS,
    ADJECTIVE_ABBREVIATION_EXPANSIONS,
)
from .constants import (
    ABBREV_DOTTED_PATTERN,
    ABBREV_MAX_LEN_EN,
    ABBREV_MAX_LEN_RU,
    ABBREV_MIN_LEN,
    ABBREV_SOLID_PATTERN,
    EN_LETTER_NAMES,
    KNOWN_ABBREVIATIONS,
    PERSON_INITIALS_SURNAME_PATTERN,
    PERSON_SINGLE_INITIAL_SURNAME_PATTERN,
    PERSON_SURNAME_INITIALS_PATTERN,
    PERSON_SURNAME_SINGLE_INITIAL_PATTERN,
    REAL_WORD_POS,
    RU_LETTER_NAMES,
    RU_VOWELS,
)
from .numerals._helpers import safe_inflect
from .options import NormalizeOptions

_ETC_ABBREVIATION_PATTERN = re.compile(
    r"(?<!\w)(?P<abbr>т\.?\s*[дп]\.)(?P<tail>\s*)(?!\w)", re.IGNORECASE
)
_ADJECTIVE_ABBREVIATION_WORD = (
    r"(?!и\b|или\b|либо\b|да\b|но\b|а\b)[А-ЯЁа-яё-]+"
)
_ADJECTIVE_ABBREVIATION_PATTERN = re.compile(
    rf"(?<!\w)(?P<abbr>{'|'.join(sorted((re.escape(key) for key in ADJECTIVE_ABBREVIATION_EXPANSIONS), key=len, reverse=True))})"
    rf"(?P<space>\s+)(?P<phrase>{_ADJECTIVE_ABBREVIATION_WORD}(?:\s+{_ADJECTIVE_ABBREVIATION_WORD}){{0,2}})",
    re.IGNORECASE,
)
_WORD_AMPERSAND_PATTERN = re.compile(
    r"(?P<left>[A-Za-zА-Яа-яЁё]+(?:[-'][A-Za-zА-Яа-яЁё0-9]+)*)\s*&\s*(?P<right>[A-Za-zА-Яа-яЁё]+(?:[-'][A-Za-zА-Яа-яЁё0-9]+)*)"
)
_SINGLE_TOKEN_EN_LETTER_NAMES = frozenset(
    value.lower() for value in EN_LETTER_NAMES.values() if " " not in value
)
_LANGUAGE_ORIGIN_ABBREVIATIONS = {
    "англ.": "английский",
    "руск.": "русский",
    "рус.": "русский",
    "немецк.": "немецкий",
    "нем.": "немецкий",
    "франц.": "французский",
    "греч.": "греческий",
    "латин.": "латинский",
    "лат.": "латинский",
}
_LANGUAGE_ORIGIN_PATTERN = re.compile(
    rf"(?<!\w)(?P<prep>от|из|с|со)\s+(?P<abbr>{'|'.join(sorted((re.escape(key) for key in _LANGUAGE_ORIGIN_ABBREVIATIONS), key=len, reverse=True))})"
    r"(?P<space>\s+)(?P<word>[^\s.,;:!?()\"\[\]{}«»]+)",
    re.IGNORECASE,
)


def _expand_contextual_etc_abbreviations(text: str) -> str:
    def repl(match: re.Match[str]) -> str:
        raw_abbr = match.group("abbr")
        tail = match.group("tail")
        normalized_abbr = re.sub(r"\s+", "", raw_abbr).lower()
        expansion = "так далее" if "д" in normalized_abbr else "тому подобное"

        rest = text[match.end() :]
        next_char = rest[:1]
        stripped_rest = rest.lstrip()
        next_significant = stripped_rest[:1]
        has_linebreak_before_next = "\n" in tail or (
            "\n" in rest and rest[: len(rest) - len(stripped_rest)] != ""
        )

        keep_terminal_dot = (
            not stripped_rest
            or has_linebreak_before_next
            or (next_significant and next_significant.isupper())
        )

        if (
            next_char in '"»”)]}'
            and len(stripped_rest) > 1
            and stripped_rest[1:2].isupper()
        ):
            keep_terminal_dot = True

        return f"{expansion}{'.' if keep_terminal_dot else ''}{tail}"

    return _ETC_ABBREVIATION_PATTERN.sub(repl, text)


def _expand_contextual_adjective_abbreviations(text: str) -> str:
    morph = get_morph()

    def choose_head_noun(phrase: str):
        for word in phrase.split():
            parsed = morph.parse(word.lower())
            noun_candidate = next(
                (candidate for candidate in parsed if "NOUN" in candidate.tag),
                None,
            )
            if noun_candidate is not None:
                return noun_candidate
        return None

    def inflect_adjective(lemma: str, noun_parse) -> str:
        parsed = morph.parse(lemma)
        adjective_parse = next(
            (
                candidate
                for candidate in parsed
                if candidate.tag.POS in {"ADJF", "PRTF"}
            ),
            parsed[0] if parsed else None,
        )
        if adjective_parse is None:
            return lemma

        noun_case = "loct" if "loc2" in noun_parse.tag else noun_parse.tag.case
        if noun_case is None:
            return lemma
        target_tags = {noun_case}
        if "plur" in noun_parse.tag:
            target_tags.add("plur")
        else:
            target_tags.add("sing")
            if noun_parse.tag.gender:
                target_tags.add(noun_parse.tag.gender)
        return safe_inflect(
            adjective_parse,
            target_tags,
            fallback_word=lemma,
            pos_filter={"ADJF", "PRTF"},
        )

    def repl(match: re.Match[str]) -> str:
        abbr = match.group("abbr")
        phrase = match.group("phrase")
        if abbr.lower() in _LANGUAGE_ORIGIN_ABBREVIATIONS and phrase[:1].isupper():
            return match.group(0)
        lemma = ADJECTIVE_ABBREVIATION_EXPANSIONS.get(abbr.lower())
        if lemma is None:
            return match.group(0)
        noun_parse = choose_head_noun(phrase)
        if noun_parse is None:
            return match.group(0)
        expanded = inflect_adjective(lemma, noun_parse)
        if abbr[:1].isupper():
            expanded = expanded[:1].upper() + expanded[1:]
        return f"{expanded}{match.group('space')}{phrase}"

    return _ADJECTIVE_ABBREVIATION_PATTERN.sub(repl, text)


def _expand_language_origin_abbreviations(text: str) -> str:
    morph = get_morph()

    def has_non_cyrillic_word(word: str) -> bool:
        letters = [char for char in word if char.isalpha()]
        return bool(letters) and not any(
            ("А" <= char <= "я") or char in "Ёё" for char in letters
        )

    def inflect_language_lemma(lemma: str) -> str:
        parsed = next(
            (
                candidate
                for candidate in morph.parse(lemma)
                if candidate.tag.POS in {"ADJF", "PRTF"}
            ),
            None,
        )
        if parsed is None:
            return lemma
        return safe_inflect(
            parsed,
            {"gent", "sing", "masc"},
            fallback_word=lemma,
            pos_filter={"ADJF", "PRTF"},
        )

    def repl(match: re.Match[str]) -> str:
        word = match.group("word")
        if not has_non_cyrillic_word(word):
            return match.group(0)
        lemma = _LANGUAGE_ORIGIN_ABBREVIATIONS.get(match.group("abbr").lower())
        if lemma is None:
            return match.group(0)
        return (
            f"{match.group('prep')} {inflect_language_lemma(lemma)}"
            f"{match.group('space')}{word}"
        )

    return _LANGUAGE_ORIGIN_PATTERN.sub(repl, text)


def expand_person_initials(text: str) -> str:
    def initial_name(ch: str) -> str:
        return RU_LETTER_NAMES.get(ch.upper(), ch.lower())

    def candidate_grammemes(candidate) -> frozenset[str]:
        grammemes = getattr(candidate.tag, "grammemes", None)
        if grammemes is None:
            return frozenset()
        return frozenset(grammemes)

    def is_likely_person_name_token(token: str) -> bool:
        parsed = get_morph().parse(token)
        if not parsed:
            return True
        meaningful = [
            candidate
            for candidate in parsed
            if "PNCT" not in candidate_grammemes(candidate)
        ]
        if not meaningful:
            return True
        if any(
            marker in candidate_grammemes(candidate)
            for candidate in meaningful
            for marker in ("Surn", "Name", "Patr")
        ):
            return True
        top_candidates = meaningful[:3]
        if top_candidates and all(
            "Geox" in candidate_grammemes(candidate) for candidate in top_candidates
        ):
            return False
        if top_candidates and all(
            "Abbr" in candidate_grammemes(candidate) for candidate in top_candidates
        ):
            return False
        return True

    def choose_tail(match_end: int) -> tuple[bool, str]:
        tail = text[match_end:]
        stripped = tail.lstrip()
        if not stripped:
            return True, ""
        if stripped[0] in ".!?…":
            return True, ""
        if stripped[0].islower() or stripped[0].isdigit():
            return False, ""
        return True, ""

    def repl_surname_initials(match):
        surname = match.group("surname")
        i1 = initial_name(match.group("i1"))
        i2 = initial_name(match.group("i2"))
        is_final, non_final_tail = choose_tail(match.end())
        if is_final:
            terminal = (
                ""
                if text[match.end() :].lstrip().startswith((".", "!", "?", "…"))
                else "."
            )
            return f"{surname} {i1} {i2}{terminal}"
        return f"{surname}, {i1}, {i2}{non_final_tail}"

    def repl_surname_single_initial(match):
        surname = match.group("surname")
        if not is_likely_person_name_token(surname):
            return match.group(0)
        i1 = initial_name(match.group("i1"))
        is_final, non_final_tail = choose_tail(match.end())
        if is_final:
            terminal = (
                ""
                if text[match.end() :].lstrip().startswith((".", "!", "?", "…"))
                else "."
            )
            return f"{surname} {i1}{terminal}"
        return f"{surname}, {i1},{non_final_tail}"

    def repl_initials_surname(match):
        i1 = initial_name(match.group("i1"))
        i2 = initial_name(match.group("i2"))
        surname = match.group("surname")
        is_final, non_final_tail = choose_tail(match.end())
        if is_final:
            terminal = (
                ""
                if text[match.end() :].lstrip().startswith((".", "!", "?", "…"))
                else "."
            )
            return f"{i1} {i2} {surname}{terminal}"
        return f"{i1} {i2} {surname}{non_final_tail}"

    def repl_single_initial_surname(match):
        surname = match.group("surname")
        if not is_likely_person_name_token(surname):
            return match.group(0)
        i1 = initial_name(match.group("i1"))
        is_final, non_final_tail = choose_tail(match.end())
        if is_final:
            terminal = (
                ""
                if text[match.end() :].lstrip().startswith((".", "!", "?", "…"))
                else "."
            )
            return f"{i1} {surname}{terminal}"
        return f"{i1} {surname}{non_final_tail}"

    text = PERSON_SURNAME_INITIALS_PATTERN.sub(repl_surname_initials, text)
    text = PERSON_INITIALS_SURNAME_PATTERN.sub(repl_initials_surname, text)
    text = PERSON_SURNAME_SINGLE_INITIAL_PATTERN.sub(repl_surname_single_initial, text)
    return PERSON_SINGLE_INITIAL_SURNAME_PATTERN.sub(repl_single_initial_surname, text)


def expand_letter_abbreviations(text: str) -> str:
    def expand_token(token: str, is_dotted: bool) -> str:
        tail_dot = token.endswith(".")
        letters = [char for char in token if char.isalpha()]
        if len(letters) < ABBREV_MIN_LEN:
            return token

        letters_upper = [char.upper() for char in letters]
        joined_key = "".join(letters_upper)
        if joined_key in KNOWN_ABBREVIATIONS:
            custom_reading = KNOWN_ABBREVIATIONS[joined_key]
            return (
                custom_reading + "."
                if custom_reading and tail_dot
                else custom_reading or token
            )

        if all(char in RU_LETTER_NAMES for char in letters_upper):
            if not is_dotted:
                parsed = get_morph().parse(joined_key.lower())[0]
                if parsed.tag.POS in REAL_WORD_POS and "Abbr" not in parsed.tag:
                    return token
            if not is_dotted:
                if len(letters_upper) > ABBREV_MAX_LEN_RU:
                    return token
                if sum(1 for char in letters_upper if char in RU_VOWELS) > 1:
                    return token
            names = [RU_LETTER_NAMES[char] for char in letters_upper]
        elif all(char in EN_LETTER_NAMES for char in letters_upper):
            if not is_dotted and len(letters_upper) > ABBREV_MAX_LEN_EN:
                return token
            names = [EN_LETTER_NAMES[char] for char in letters_upper]
        else:
            return token

        expanded = " ".join(names)
        return expanded + "." if tail_dot else expanded

    text = ABBREV_DOTTED_PATTERN.sub(lambda m: expand_token(m.group(0), True), text)
    return ABBREV_SOLID_PATTERN.sub(lambda m: expand_token(m.group(0), False), text)


def _normalize_contextual_ampersands(
    text: str, options: NormalizeOptions | None = None
) -> str:
    active = options or NormalizeOptions()

    def has_ascii_letters(token: str) -> bool:
        return any("A" <= char <= "Z" or "a" <= char <= "z" for char in token)

    def has_cyrillic_letters(token: str) -> bool:
        return any(("А" <= char <= "я") or char in "Ёё" for char in token)

    def is_english_letter_name(token: str) -> bool:
        return token.lower() in _SINGLE_TOKEN_EN_LETTER_NAMES

    def repl(match: re.Match[str]) -> str:
        left = match.group("left")
        right = match.group("right")
        left_has_ascii = has_ascii_letters(left)
        right_has_ascii = has_ascii_letters(right)
        left_is_english = is_english_letter_name(left)
        right_is_english = is_english_letter_name(right)

        brand_like_context = (
            (left_has_ascii and right_has_ascii and active.enable_latinization)
            or (left_is_english and right_is_english)
            or (left_is_english and right_has_ascii)
            or (left_has_ascii and right_is_english)
        )
        if brand_like_context:
            return f"{left} энд {right}"

        if left_has_ascii and right_has_ascii:
            return f"{left} & {right}"

        if (
            has_cyrillic_letters(left)
            or has_cyrillic_letters(right)
            or left_is_english
            or right_is_english
        ):
            return f"{left} и {right}"

        return match.group(0)

    return _WORD_AMPERSAND_PATTERN.sub(repl, text)


def expand_abbreviations(text: str, options: NormalizeOptions | None = None) -> str:
    active = options or NormalizeOptions()
    if not active.enable_abbreviation_expansion:
        return text
    text = _expand_language_origin_abbreviations(text)
    if active.enable_contextual_abbreviation_expansion:
        text = _expand_contextual_etc_abbreviations(text)
        text = _expand_contextual_adjective_abbreviations(text)
        for pattern, replacement in ABBREVIATION_PATTERNS:
            text = pattern.sub(replacement, text)
    if active.enable_initials_expansion:
        text = expand_person_initials(text)
    if active.enable_letter_abbreviation_expansion:
        text = expand_letter_abbreviations(text)
    return _normalize_contextual_ampersands(text, active)



from __future__ import annotations

from ..options import NormalizeOptions
from ..text_context import simple_tokenize
from ._constants import ALL_UNITS
from ._helpers import get_numeral_case
from .cardinals import (
    normalize_all_digits_everywhere,
    normalize_cardinal_numerals,
    normalize_numeric_unit_ranges,
    normalize_remaining_post_numeral_abbreviations,
)
from .decimals import normalize_decimals
from .fractions import normalize_fractions
from .ordinals import (
    normalize_heading_numbers,
    normalize_heading_ranges,
    normalize_hyphenated_words,
    normalize_ordinals,
)
from .symbols import (
    normalize_greek_letters,
    normalize_math_symbols,
    normalize_standalone_currency,
)


def normalize_numerals(text: str, options: NormalizeOptions | None = None) -> str:
    del options
    text = normalize_heading_ranges(text)
    text = normalize_heading_numbers(text)
    text = normalize_numeric_unit_ranges(text)
    text = normalize_math_symbols(text)
    text = normalize_cardinal_numerals(text)
    text = normalize_remaining_post_numeral_abbreviations(text)
    text = normalize_greek_letters(text)
    text = normalize_standalone_currency(text)
    text = normalize_all_digits_everywhere(text)
    return text


__all__ = [
    "ALL_UNITS",
    "get_numeral_case",
    "normalize_all_digits_everywhere",
    "normalize_cardinal_numerals",
    "normalize_decimals",
    "normalize_fractions",
    "normalize_greek_letters",
    "normalize_heading_numbers",
    "normalize_heading_ranges",
    "normalize_hyphenated_words",
    "normalize_math_symbols",
    "normalize_numerals",
    "normalize_numeric_unit_ranges",
    "normalize_ordinals",
    "normalize_remaining_post_numeral_abbreviations",
    "normalize_standalone_currency",
    "simple_tokenize",
]



from __future__ import annotations

import re

import num2words

from .options import NormalizeOptions
from .years import year_to_ordinal_words

MONTHS_GENT = {
    "января": 1,
    "февраля": 2,
    "марта": 3,
    "апреля": 4,
    "мая": 5,
    "июня": 6,
    "июля": 7,
    "августа": 8,
    "сентября": 9,
    "октября": 10,
    "ноября": 11,
    "декабря": 12,
}
TEXT_DATE_RANGE_CASES = {
    "за": "accusative",
    "на": "accusative",
    "в": "accusative",
    "во": "accusative",
    "по": "accusative",
    "с": "genitive",
    "со": "genitive",
    "до": "genitive",
    "от": "genitive",
}
TEXT_DATE_RANGE_PATTERN = re.compile(
    r"\b(?:(?P<prep>за|на|в|во|по|с|со|до|от)\s+)?(?P<day1>\d{1,2})\s*[–—-]\s*(?P<day2>\d{1,2})\s+(?P<month>"
    + "|".join(MONTHS_GENT.keys())
    + r")(?:\s+(?P<year>\d{4})\s*(?:года?)?)?",
    re.IGNORECASE,
)
TEXT_DATE_LIST_PATTERN = re.compile(
    r"\b(?P<days>\d{1,2}(?:\s*(?:,|и)\s*\d{1,2})+)\s+(?P<month>"
    + "|".join(MONTHS_GENT.keys())
    + r")(?:\s+(?P<year>\d{4})\s*(?:года?)?)?",
    re.IGNORECASE,
)
TEXT_DATE_PATTERN = re.compile(
    r"\b(\d{1,2})\s+("
    + "|".join(MONTHS_GENT.keys())
    + r")(?:\s+(\d{4})\s*(?:года?)?)?",
    re.IGNORECASE,
)
MONTHS_GENT_BY_NUMBER = {
    1: "января",
    2: "февраля",
    3: "марта",
    4: "апреля",
    5: "мая",
    6: "июня",
    7: "июля",
    8: "августа",
    9: "сентября",
    10: "октября",
    11: "ноября",
    12: "декабря",
}
NUMERIC_DATE_PATTERN = re.compile(
    r"\b(?P<day>0?[1-9]|[12]\d|3[01])\.(?P<month>0?[1-9]|1[0-2])\.(?P<year>\d{2,4})\b"
)
TIME_PATTERN = re.compile(r"\b(?P<hour>[01]?\d|2[0-3]):(?P<minute>[0-5]\d)\b")
DOTTED_TIME_PATTERN = re.compile(
    r"\b(?P<hour>[01]?\d|2[0-3])\.(?P<minute>[0-5]\d)\b",
    re.IGNORECASE,
)
DOTTED_TIME_LEFT_CONTEXT_PATTERN = re.compile(
    r"(?:^|[\s(\"«])(?:в|во|к|около)\s*$",
    re.IGNORECASE,
)
DOTTED_TIME_RIGHT_CONTEXT_PATTERN = re.compile(
    r"^\s*(?:утр(?:ом|а)?|вечер(?:ом|а)?|ноч(?:ью|и)?|дн(?:ём|ем)?|час(?:ов|а)?\b)",
    re.IGNORECASE,
)


def _day_to_ordinal_genitive(day: int) -> str | None:
    if day < 1 or day > 31:
        return None
    try:
        return num2words.num2words(
            day, lang="ru", to="ordinal", case="genitive", gender="n"
        )
    except Exception:
        try:
            return num2words.num2words(day, lang="ru", to="ordinal")
        except Exception:
            return None


def _day_to_ordinal(day: int, case: str = "genitive") -> str | None:
    if day < 1 or day > 31:
        return None
    try:
        return num2words.num2words(day, lang="ru", to="ordinal", case=case, gender="n")
    except Exception:
        try:
            return num2words.num2words(day, lang="ru", to="ordinal")
        except Exception:
            return None


def normalize_text_dates(text: str) -> str:
    def range_repl(match: re.Match[str]) -> str:
        prep = match.group("prep")
        case = (
            TEXT_DATE_RANGE_CASES.get(prep.lower(), "genitive") if prep else "genitive"
        )
        day1_word = _day_to_ordinal(int(match.group("day1")), case=case)
        day2_word = _day_to_ordinal(int(match.group("day2")), case=case)
        month = match.group("month").lower()
        if day1_word is None or day2_word is None:
            return match.group(0)
        result = f"{day1_word} — {day2_word} {month}"
        if prep:
            result = f"{prep} {result}"
        if match.group("year"):
            result += (
                f" {year_to_ordinal_words(int(match.group('year')), case='gent')} года"
            )
        return result

    def list_repl(match: re.Match[str]) -> str:
        month = match.group("month").lower()
        parts = re.split(r"(\s*(?:,|и)\s*)", match.group("days"))
        rendered_parts: list[str] = []
        for part in parts:
            stripped = part.strip()
            if not stripped:
                continue
            if stripped in {",", "и"}:
                rendered_parts.append(part)
                continue
            day_word = _day_to_ordinal_genitive(int(stripped))
            if day_word is None:
                return match.group(0)
            rendered_parts.append(day_word)
        result = f"{''.join(rendered_parts)} {month}"
        if match.group("year"):
            result += (
                f" {year_to_ordinal_words(int(match.group('year')), case='gent')} года"
            )
        return result

    def repl(match: re.Match[str]) -> str:
        day_word = _day_to_ordinal_genitive(int(match.group(1)))
        month = match.group(2).lower()
        if day_word is None:
            return match.group(0)
        result = f"{day_word} {month}"
        if match.group(3):
            result += f" {year_to_ordinal_words(int(match.group(3)), case='gent')} года"
        return result

    text = TEXT_DATE_RANGE_PATTERN.sub(range_repl, text)
    text = TEXT_DATE_LIST_PATTERN.sub(list_repl, text)
    return TEXT_DATE_PATTERN.sub(repl, text)


def normalize_dates(text: str) -> str:
    def repl(match: re.Match[str]) -> str:
        day = int(match.group("day"))
        month = int(match.group("month"))
        year = int(match.group("year"))
        if day < 1 or day > 31 or month < 1 or month > 12:
            return match.group(0)
        if year < 100:
            year = 1900 + year if year >= 50 else 2000 + year
        try:
            day_words = num2words.num2words(
                day, lang="ru", to="ordinal", case="genitive", gender="n"
            )
        except Exception:
            day_words = num2words.num2words(day, lang="ru", to="ordinal")
        try:
            year_words = num2words.num2words(
                year, lang="ru", to="ordinal", case="genitive"
            )
        except Exception:
            year_words = num2words.num2words(year, lang="ru", to="ordinal")
        return f"{day_words} {MONTHS_GENT_BY_NUMBER.get(month, str(month))} {year_words} года"

    return NUMERIC_DATE_PATTERN.sub(repl, text)


def normalize_time(text: str) -> str:
    def render_time(match: re.Match[str]) -> str:
        hour = int(match.group("hour"))
        minute_str = match.group("minute")
        try:
            hour_words = num2words.num2words(hour, lang="ru")
        except Exception:
            hour_words = str(hour)
        if minute_str[0] == "0":
            try:
                minute_words = (
                    f"ноль {num2words.num2words(int(minute_str[1]), lang='ru')}"
                )
            except Exception:
                minute_words = f"ноль {minute_str[1]}"
        else:
            try:
                minute_words = num2words.num2words(int(minute_str), lang="ru")
            except Exception:
                minute_words = minute_str
        return f"{hour_words}, {minute_words}"

    def render_dotted_time(match: re.Match[str]) -> str:
        left_context = text[max(0, match.start() - 16) : match.start()]
        right_context = text[match.end() : match.end() + 24]
        if not (
            DOTTED_TIME_LEFT_CONTEXT_PATTERN.search(left_context)
            or DOTTED_TIME_RIGHT_CONTEXT_PATTERN.search(right_context)
        ):
            return match.group(0)
        return render_time(match)

    text = TIME_PATTERN.sub(render_time, text)
    return DOTTED_TIME_PATTERN.sub(render_dotted_time, text)


def normalize_dates_and_time(text: str, options: NormalizeOptions | None = None) -> str:
    active = options or NormalizeOptions()
    if not active.enable_dates_time_normalization:
        return text
    text = normalize_text_dates(text)
    text = normalize_dates(text)
    text = normalize_time(text)
    return text



from __future__ import annotations

import num2words

from ._morph import get_morph

CASE_TO_NUM2WORDS = {
    "nomn": "nominative",
    "gent": "genitive",
    "datv": "dative",
    "accs": "accusative",
    "ablt": "instrumental",
    "loct": "prepositional",
}
GENDER_TO_NUM2WORDS = {"masc": "m", "femn": "f", "neut": "n"}


def noun_parse_case(noun_parse) -> str:
    return "loct" if "loc2" in noun_parse.tag else (noun_parse.tag.case or "nomn")


def choose_noun_parse(word: str, prefer_inanimate: bool = True):
    noun_parses = [
        candidate for candidate in get_morph().parse(word.lower()) if "NOUN" in candidate.tag
    ]
    if not noun_parses:
        return None
    if not prefer_inanimate:
        return noun_parses[0]
    inanimate_parses = [candidate for candidate in noun_parses if "inan" in candidate.tag]
    return inanimate_parses[0] if inanimate_parses else noun_parses[0]


def render_ordinal(
    number: int,
    case: str = "nomn",
    gender: str | None = None,
    plural: bool = False,
    inanimate: bool = False,
) -> str:
    generation_case = case
    if case == "accs" and not plural and inanimate and gender in {"masc", "neut"}:
        generation_case = "nomn"

    kwargs: dict[str, str | bool] = {
        "lang": "ru",
        "to": "ordinal",
        "case": CASE_TO_NUM2WORDS.get(generation_case, "nominative"),
    }
    if plural:
        kwargs["plural"] = True
    elif gender in GENDER_TO_NUM2WORDS:
        kwargs["gender"] = GENDER_TO_NUM2WORDS[gender]

    try:
        return num2words.num2words(number, **kwargs)
    except Exception:
        pass

    try:
        ordinal = num2words.num2words(number, lang="ru", to="ordinal")
    except Exception:
        return str(number)

    words = ordinal.split()
    if not words:
        return ordinal

    parsed = get_morph().parse(words[-1])
    if not parsed:
        return ordinal

    target_tags = {generation_case}
    if plural:
        target_tags.add("plur")
    elif gender:
        target_tags.add(gender)
    inflected = parsed[0].inflect(target_tags)
    if inflected:
        words[-1] = inflected.word
    return " ".join(words)


def render_ordinal_from_noun_word(
    number: int,
    noun_word: str,
    *,
    prefer_inanimate: bool = True,
    singularize_plural: bool = False,
) -> str | None:
    noun_parse = choose_noun_parse(noun_word, prefer_inanimate=prefer_inanimate)
    if noun_parse is None:
        return None
    return render_ordinal_from_noun_parse(
        number,
        noun_parse,
        singularize_plural=singularize_plural,
    )


def render_ordinal_from_noun_parse(
    number: int,
    noun_parse,
    *,
    singularize_plural: bool = False,
) -> str:
    case = noun_parse_case(noun_parse)
    gender = noun_parse.tag.gender or "masc"
    plural = "plur" in noun_parse.tag and not singularize_plural
    inanimate = "inan" in noun_parse.tag
    return render_ordinal(
        number,
        case=case,
        gender=gender,
        plural=plural,
        inanimate=inanimate,
    )



from __future__ import annotations

import re
from typing import Literal

from ..preprocess_utils import NEGATIVE_NUMBER_PLACEHOLDER
from ._constants import UNIT_TOKEN_FRAGMENT, UNITS_DATA

ORDINAL_SUFFIXES = {"й", "я", "е", "го", "му", "м", "ю", "ее", "ий", "ая", "ое"}
CARDINAL_CASE_SUFFIXES = {"ти", "ми", "х", "мя", "и"}
AMBIGUOUS_SINGLE_LETTER_HYPHEN_UNITS = {"г", "л", "м", "н", "р", "с", "т", "ф", "ч"}
NUMERIC_UNIT_HYPHEN_PATTERN = re.compile(
    rf"(?<![\d{re.escape(NEGATIVE_NUMBER_PLACEHOLDER)}])"
    rf"(?P<num>(?:{re.escape(NEGATIVE_NUMBER_PLACEHOLDER)}|-)?\d+(?:[.,]\d+)?)"
    rf"\s*[-–—]\s*"
    rf"(?P<unit>{UNIT_TOKEN_FRAGMENT})(?P<unit_dot>\.)?"
    rf"(?!\w)",
    re.UNICODE,
)
SPACED_NUMERIC_HYPHEN_WORD_PATTERN = re.compile(
    r"(?<!\d)(?P<num>\d+)\s+-\s+(?P<rhs>[а-яА-ЯёЁ]{1,})(?!\w)",
    re.UNICODE,
)

NumericHyphenKind = Literal["unit", "ordinal_suffix", "cardinal_case_suffix", "word"]


def is_safe_numeric_hyphen_unit(unit_raw: str) -> bool:
    unit_key = unit_raw.lower().strip(".")
    if unit_key not in UNITS_DATA:
        return False
    if len(unit_key) == 1 and unit_key in AMBIGUOUS_SINGLE_LETTER_HYPHEN_UNITS:
        return False
    return True


def classify_numeric_hyphen_rhs(word: str) -> NumericHyphenKind:
    word_lower = word.lower().strip(".")
    if is_safe_numeric_hyphen_unit(word):
        return "unit"
    if word_lower in ORDINAL_SUFFIXES:
        return "ordinal_suffix"
    if word_lower in CARDINAL_CASE_SUFFIXES:
        return "cardinal_case_suffix"
    return "word"


def normalize_numeric_unit_hyphen_links(text: str) -> str:
    def repl(match: re.Match[str]) -> str:
        unit = match.group("unit")
        unit_dot = match.group("unit_dot") or ""
        if not is_safe_numeric_hyphen_unit(unit + unit_dot):
            return match.group(0)
        return f"{match.group('num')} {unit}{unit_dot}"

    return NUMERIC_UNIT_HYPHEN_PATTERN.sub(repl, text)


def normalize_spaced_numeric_hyphen_words(text: str) -> str:
    def repl(match: re.Match[str]) -> str:
        return f"{match.group('num')}-{match.group('rhs')}"

    return SPACED_NUMERIC_HYPHEN_WORD_PATTERN.sub(repl, text)



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path
from typing import Literal, Optional

NormalizeMode = Literal["safe", "tts"]
LatinizationBackend = Literal["ipa", "dictionary"]

_SAFE_MODE_DEFAULTS = {
    "enable_caps_normalization": False,
    "enable_first_word_decap": False,
    "remove_links": False,
    "enable_url_normalization": False,
    "enable_year_normalization": True,
    "enable_roman_normalization": True,
    "enable_dates_time_normalization": True,
    "enable_numeral_normalization": True,
    "enable_abbreviation_expansion": True,
    "enable_contextual_abbreviation_expansion": True,
    "enable_initials_expansion": False,
    "enable_letter_abbreviation_expansion": False,
    "enable_dictionary_normalization": True,
    "enable_latinization": False,
    "latinization_backend": "ipa",
    "enable_latinization_stress_marks": False,
}

_TTS_MODE_DEFAULTS = {
    "enable_caps_normalization": True,
    "enable_first_word_decap": True,
    "remove_links": True,
    "enable_url_normalization": True,
    "enable_year_normalization": True,
    "enable_roman_normalization": True,
    "enable_dates_time_normalization": True,
    "enable_numeral_normalization": True,
    "enable_abbreviation_expansion": True,
    "enable_contextual_abbreviation_expansion": True,
    "enable_initials_expansion": True,
    "enable_letter_abbreviation_expansion": True,
    "enable_dictionary_normalization": True,
    "enable_latinization": True,
    "latinization_backend": "ipa",
    "enable_latinization_stress_marks": False,
}


def _mode_defaults(mode: NormalizeMode | None) -> dict[str, object]:
    if mode == "tts":
        return dict(_TTS_MODE_DEFAULTS)
    return dict(_SAFE_MODE_DEFAULTS)


@dataclass(frozen=True, slots=True, init=False)
class NormalizeOptions:
    """Public configuration for ru_normalizr."""

    enable_caps_normalization: bool
    enable_first_word_decap: bool
    remove_links: bool
    enable_url_normalization: bool
    remove_links_ignore_interval: tuple[int, int]
    enable_year_normalization: bool
    enable_roman_normalization: bool
    enable_dates_time_normalization: bool
    enable_numeral_normalization: bool
    enable_abbreviation_expansion: bool
    enable_contextual_abbreviation_expansion: bool
    enable_initials_expansion: bool
    enable_letter_abbreviation_expansion: bool
    enable_dictionary_normalization: bool
    enable_latinization: bool
    latinization_backend: LatinizationBackend
    enable_latinization_stress_marks: bool
    latin_dictionary_filename: str
    dictionary_include_files: tuple[str, ...]
    dictionary_exclude_files: tuple[str, ...]
    dictionaries_path: Optional[Path]
    mode: NormalizeMode | None

    def __init__(
        self,
        *,
        enable_caps_normalization: bool | None = None,
        enable_first_word_decap: bool | None = None,
        remove_links: bool | None = None,
        enable_url_normalization: bool | None = None,
        remove_links_ignore_interval: tuple[int, int] = (1000, 2200),
        enable_year_normalization: bool | None = None,
        enable_roman_normalization: bool | None = None,
        enable_dates_time_normalization: bool | None = None,
        enable_numeral_normalization: bool | None = None,
        enable_abbreviation_expansion: bool | None = None,
        enable_contextual_abbreviation_expansion: bool | None = None,
        enable_initials_expansion: bool | None = None,
        enable_letter_abbreviation_expansion: bool | None = None,
        enable_dictionary_normalization: bool | None = None,
        enable_latinization: bool | None = None,
        latinization_backend: LatinizationBackend | None = None,
        enable_latinization_stress_marks: bool | None = None,
        latin_dictionary_filename: str = "latinization_rules.dic",
        dictionary_include_files: tuple[str, ...] = (),
        dictionary_exclude_files: tuple[str, ...] = (),
        dictionaries_path: Optional[Path] = None,
        mode: NormalizeMode | None = None,
    ) -> None:
        defaults = _mode_defaults(mode)

        values = {
            "enable_caps_normalization": enable_caps_normalization,
            "enable_first_word_decap": enable_first_word_decap,
            "remove_links": remove_links,
            "enable_url_normalization": enable_url_normalization,
            "enable_year_normalization": enable_year_normalization,
            "enable_roman_normalization": enable_roman_normalization,
            "enable_dates_time_normalization": enable_dates_time_normalization,
            "enable_numeral_normalization": enable_numeral_normalization,
            "enable_abbreviation_expansion": enable_abbreviation_expansion,
            "enable_contextual_abbreviation_expansion": enable_contextual_abbreviation_expansion,
            "enable_initials_expansion": enable_initials_expansion,
            "enable_letter_abbreviation_expansion": enable_letter_abbreviation_expansion,
            "enable_dictionary_normalization": enable_dictionary_normalization,
            "enable_latinization": enable_latinization,
            "latinization_backend": latinization_backend,
            "enable_latinization_stress_marks": enable_latinization_stress_marks,
        }

        for field_name, explicit_value in values.items():
            object.__setattr__(
                self,
                field_name,
                defaults[field_name] if explicit_value is None else explicit_value,
            )

        object.__setattr__(self, "remove_links_ignore_interval", remove_links_ignore_interval)
        object.__setattr__(self, "latin_dictionary_filename", latin_dictionary_filename)
        object.__setattr__(self, "dictionary_include_files", dictionary_include_files)
        object.__setattr__(self, "dictionary_exclude_files", dictionary_exclude_files)
        object.__setattr__(self, "dictionaries_path", dictionaries_path)
        object.__setattr__(self, "mode", mode)

    @classmethod
    def safe(cls, **overrides: object) -> "NormalizeOptions":
        return cls(mode="safe", **overrides)

    @classmethod
    def tts(cls, **overrides: object) -> "NormalizeOptions":
        return cls(mode="tts", **overrides)



from __future__ import annotations

from collections.abc import Iterable

from .normalizer import PipelineNormalizer, preprocess_text
from .options import NormalizeOptions


class Normalizer:
    """Normalization-only facade for Russian text."""

    def __init__(self, options: NormalizeOptions | None = None) -> None:
        self.options = options or NormalizeOptions()
        self._normalizer = PipelineNormalizer(self.options)

    def normalize(self, text: str) -> str:
        return self._normalizer.normalize_text(text)

    def normalize_batch(self, texts: Iterable[str]) -> list[str]:
        return [self.normalize(text) for text in texts]

    def run_stage(self, stage: str, text: str) -> str:
        return self._normalizer.run_stage(stage, text)


def normalize(text: str, options: NormalizeOptions | None = None) -> str:
    return Normalizer(options).normalize(text)


__all__ = ["Normalizer", "normalize", "preprocess_text"]
